"""
Compile the monster bibfile from the BibTeX files listed in references/BIBFILES.ini
"""
from pyglottolog.monster import compile


def register(parser):
    parser.add_argument(
        '--rebuild',
        default=False,
        action='store_true',
    )


def run(args):
    compile(args.repos, args.log, rebuild=args.rebuild)



import io
import json
import zipfile
import itertools
import collections

from csvw.dsv import reader
import requests
from pycldf.dataset import MD_SUFFIX


def read_cldf_languages(url):  # pragma: no cover
    r = requests.get(url)
    with zipfile.ZipFile(io.BytesIO(r.content)) as zip:
        for member in zip.namelist():
            if member.endswith(MD_SUFFIX):
                break
        else:
            raise ValueError('No metadata file found')

        with zip.open(member) as fp:
            md = json.loads(fp.read().decode('utf8'))

        for table in md['tables']:
            if table.get('dc:conformsTo') == 'http://cldf.clld.org/v1.0/terms.rdf#LanguageTable':
                lurl, schema = table['url'], table['tableSchema']['columns']
                break
        else:
            raise ValueError('No LanguageTable found')

        for member in zip.namelist():
            if member.endswith(lurl):
                with zip.open(member) as fp:
                    return reader([line.strip() for line in fp.readlines()], dicts=True), schema
        else:
            raise ValueError('LanguageTable url not found in zip')


def read_grouped_cldf_languages(url):  # pragma: no cover
    langs, schema = read_cldf_languages(url)
    colmap, gccol = {}, None
    for col in schema:
        colmap[col['name']] = col.get('propertyUrl', col['name'])
        if col.get('propertyUrl') == 'http://cldf.clld.org/v1.0/terms.rdf#glottocode':
            gccol = col['name']

    if not gccol:
        raise ValueError('No glottocode column in LanguageTable')

    for gc, langs in itertools.groupby(sorted(langs, key=lambda d: d[gccol]), lambda d: d[gccol]):
        yield gc, [{colmap[k]: v for k, v in lang.items()} for lang in langs]


class LinkProvider(object):
    __domain__ = None
    __cldf_dataset_url__ = None
    __url_template__ = None
    __label_template__ = None

    def iterupdated(self, languoids):  # pragma: no cover
        if self.__domain__ and self.__cldf_dataset_url__ and self.__url_template__:
            # FIXME: Ideally, we'd want the CLDF data to specify full URLs for languages via a
            # valueUrl property on the ID column (or the LanguageTable?).
            res = collections.defaultdict(list)
            for gc, langs in read_grouped_cldf_languages(self.__cldf_dataset_url__):
                for lang in langs:
                    lang = {k: v.strip() if isinstance(v, str) else v for k, v in lang.items()}
                    item = self.__url_template__.format(lang)
                    if self.__label_template__:
                        item = (item, self.__label_template__.format(lang))
                    res[gc].append(item)
            for lang in languoids:
                if lang.update_links(self.__domain__, res.get(lang.id, [])):
                    yield lang
        else:
            raise NotImplementedError()


class PHOIBLE(LinkProvider):  # pragma: no cover
    __domain__ = 'phoible.org'

    def __init__(self):
        r = requests.get('https://doi.org/10.5281/zenodo.2562766')
        record_id = r.url.split("/")[-1]
        r = requests.get('https://zenodo.org/api/records/%s' % record_id)
        self.__cldf_dataset_url__ = r.json()['files'][0]['links']['self']

    def iterupdated(self, languoids):  # pragma: no cover
        urls = {}
        for gc, langs in read_grouped_cldf_languages(self.__cldf_dataset_url__):
            urls[gc] = ['https://{0}/languages/{1}'.format(self.__domain__, gc)]
        for lang in languoids:
            if lang.update_links(self.__domain__, urls.get(lang.id, [])):
                yield lang


class APICS(LinkProvider):  # pragma: no cover
    __domain__ = "apics-online.info"
    __cldf_dataset_url__ = "https://cdstar.shh.mpg.de/bitstreams/" \
                           "EAEA0-61A2-213C-A9D8-0/apics_dataset.cldf.zip"
    __url_template__ = 'https://' + __domain__ + \
                       '/contributions/{0[http://cldf.clld.org/v1.0/terms.rdf#id]}'
    __label_template__ = '{0[http://cldf.clld.org/v1.0/terms.rdf#name]}'


class WALS(LinkProvider):  # pragma: no cover
    __domain__ = "wals.info"
    __cldf_dataset_url__ = "https://cdstar.shh.mpg.de/bitstreams/" \
                           "EAEA0-7269-77E5-3E10-0/wals_dataset.cldf.zip"
    __url_template__ = 'https://' + __domain__ + \
                       '/languoid/lect/wals_code_{0[http://cldf.clld.org/v1.0/terms.rdf#id]}'
    __label_template__ = '{0[http://cldf.clld.org/v1.0/terms.rdf#name]}'



import re
import pathlib
import tempfile
import subprocess

import requests
from clldutils.source import Source
try:
    from bs4 import BeautifulSoup as bs  # noqa: N813
except ImportError:  # pragma: no cover
    bs = None

LS_INDEX = 'http://www.elpublishing.org/language-snapshots'

LC_INDEX = 'http://www.elpublishing.org/language-contexts'


def download(bibfile, log):  # pragma: no cover
    fname = scrape(pathlib.Path(tempfile.gettempdir()) / 'elpub.bib')
    bibfile.update(fname, log=log)
    bibfile.check(log)
    fname.unlink()


def get(url):  # pragma: no cover
    # Some user agents seem to be rejected
    return requests.get(url, headers={'User-Agent': 'scrapy'})


def get_html(url):  # pragma: no cover
    return bs(get(url).text, 'html5lib')


def get_language_name(title):  # pragma: no cover
    if 'Contexts' in title and (':' in title):
        return title.split(':', maxsplit=1)[-1].strip()
    return title.split(')')[0] + ')'


def language_code_from_pdf(url):  # pragma: no cover
    """
    Parse the associated language code from the PDF.
    """
    iso = None

    with tempfile.TemporaryDirectory() as d:
        pdf = pathlib.Path(d) / url.split('/')[-1]
        txt = pathlib.Path(d) / url.split('/')[-1].replace('.pdf', '.txt')
        # Download the PDF
        with pdf.open('wb') as f:
            f.write(get(url).content)
        # Run pdftotext
        subprocess.check_call(
            ['pdftotext', str(pdf), str(txt)],
            stderr=subprocess.DEVNULL,
        )
        # Identify the code - which is assumed to appear in a table,
        # which is extracted as "one cell per line" so can be identified
        # by regex applied to the complete line.
        for line in txt.read_text(encoding='utf8').split('\n'):
            line = line.strip()
            if re.match('[a-z]{4}[0-9]{4}$', line):
                return line
            m = re.match(r'Glottolog Code:\s*(?P<code>[a-z]{4}[0-9]{4})', line)
            if m:
                return m.group('code')
            if re.match('([a-z]{3}|[A-Z]{3})$', line):
                iso = line.lower()
            m = re.match(r'ISO 639-3 Code:\s*(?P<code>[a-z]{3}|[A-Z]{3})$', line)
            if m:
                iso = m.group('code')
    return iso


def scrape_article(url, hhtype):  # pragma: no cover
    html = get_html(url)
    md = {
        'title': html.find('h3').text,
        'author': [],
        'hhtype': hhtype,
        'journal': 'Language Documentation and Description',
        'url': url,
    }
    pdf_url = None
    for div in html.find_all('div'):
        if div.text.startswith('Link to item:'):
            pdf_url = div.find('a')['href']
            assert pdf_url.endswith('.pdf')
            code = language_code_from_pdf(pdf_url)
            if code:
                md['lgcode'] = '{} [{}]'.format(get_language_name(md['title']), code)
        if div.find('span') and div.find('span').text.startswith('Pages'):
            md['pages'] = div.find('div').text
        if div.text.startswith('Date: '):
            md['year'] = div.text.split(':')[1].strip()
    for td in html.find_all('td'):
        link = td.find('a')
        if link and link.attrs.get('href').startswith('/authorpage'):
            md['author'].append(link.text)
    assert pdf_url
    match = re.search(r'/ldd(?P<volume>[0-9]+)_[0-9]+\.pdf', pdf_url)
    md.update(match.groupdict())
    md['author'] = ' and '.join(md['author'])
    return Source('article', url.split('/')[-1], **md)


def scrape(fname):  # pragma: no cover
    records = []
    index = get_html(LS_INDEX)
    for td in index.find_all('td'):
        if 'title' in td.attrs.get('class', []):
            link = td.find('a')
            records.append(scrape_article(link['href'], 'overview'))
    index = get_html(LC_INDEX)
    for link in index.find_all('a'):
        if '/itempage/' in link.attrs.get('href', ''):
            records.append(scrape_article(link['href'], 'overview'))
    fname.write_text(
        '\n'.join([r.bibtex() for r in sorted(records, key=lambda r: int(r.id))]), encoding='utf8')
    return fname



"""Load references from .bib files into sqlite3, hash, assign ids (split/merge)."""

import collections
import contextlib
import difflib
import functools
import itertools
import logging
import operator
import pathlib
import typing

from clldutils import jsonlib
from csvw import dsv
import sqlalchemy as sa
import sqlalchemy.orm

from .. import _compat
from ..util import (unique,
                    group_first as groupby_first)
from . import bibtex

__all__ = ['Database']

UNION_FIELDS = {'fn', 'asjp_name', 'isbn'}

IGNORE_FIELDS = {'crossref', 'numnote', 'glotto_id'}

REF_ID_FIELD = 'glottolog_ref_id'

ENCODING = 'utf-8'

SQLALCHEMY_FUTURE = True

PAGE_SIZE = 32_768

ENTRYTYPE = 'ENTRYTYPE'


log = logging.getLogger('pyglottolog')


def distance(left, right,
             *, weight={'author': 3, 'year': 3, 'title': 3, ENTRYTYPE: 2}):
    """Simple measure of the difference between two BibTeX field dicts."""
    if not (left or right):
        return 0.0

    keys = left.keys() & right.keys()
    if not keys:
        return 1.0

    weights = {k: weight.get(k, 1) for k in keys}
    ratios = (w * difflib.SequenceMatcher(None, left[k], right[k]).ratio()
              for k, w in weights.items())
    return 1 - (sum(ratios) / sum(weights.values()))


registry = sa.orm.registry()


class Connectable:
    """SQLite database."""

    def __init__(self, filepath,
                 *, future: bool = SQLALCHEMY_FUTURE,
                 paramstyle: typing.Optional[str] = 'qmark'):
        self.filepath = pathlib.Path(filepath)
        self._engine = sa.create_engine(f'sqlite:///{self.filepath}',
                                        future=future,
                                        paramstyle=paramstyle)

    def connect(self,
                *, pragma_bulk_insert: bool = False,
                page_size: typing.Optional[int] = None):
        """Connect to engine, optionally apply SQLite PRAGMAs, return conn."""
        conn = self._engine.connect()

        if page_size is not None:
            conn.execute(sa.text(f'PRAGMA page_size = {page_size:d}'))

        if pragma_bulk_insert:
            conn.execute(sa.text('PRAGMA synchronous = OFF'))
            conn.execute(sa.text('PRAGMA journal_mode = MEMORY'))

        return conn

    @contextlib.contextmanager
    def execute(self, statement, *, closing: bool = True):
        """Connect to engine, execute ``statement``, return ``CursorResult``, close."""
        with self.connect() as conn:
            result = conn.execute(statement)
            manager = contextlib.closing if closing else _compat.nullcontext
            with manager(result) as result:
                yield result


class BaseDatabase(Connectable):
    """Collection of parsed .bib files loaded into a SQLite database."""

    @classmethod
    def from_bibfiles(cls, bibfiles, filepath, *, rebuild: bool = False,
                      page_size: typing.Optional[int] = PAGE_SIZE,
                      verbose: bool = False):
        """Load ``bibfiles`` if needed, hash, split/merge, return the database."""
        self = cls(filepath)
        if self.filepath.exists():
            if not rebuild and self.is_uptodate(bibfiles):
                return self
            self.filepath.unlink()

        with self.connect(page_size=page_size) as conn:
            registry.metadata.create_all(conn)
            conn.commit()

        with self.connect(pragma_bulk_insert=True) as conn:
            import_bibfiles(conn, bibfiles)
            conn.commit()

            Entry.stats(conn=conn)
            Value.fieldstats(conn=conn)

            generate_hashes(conn)
            conn.commit()

            Entry.hashstats(conn=conn)
            Entry.hashidstats(conn=conn)

            assign_ids(conn, verbose=verbose)
            conn.commit()

        return self

    def is_uptodate(self, bibfiles, *, verbose: bool = False):
        """Does the db have the same filenames, sizes, and mtimes as ``bibfiles``?"""
        with self.connect() as conn:
            return File.same_as(conn, bibfiles, verbose=verbose)

    def __iter__(self, *, chunksize: int = 100):
        """Yield pairs of ``(Entry.id, Entry.hash)`` and unmerged field values."""
        with self.connect() as conn:
            assert Entry.allid(conn=conn)
            assert Entry.onetoone(conn=conn)

        select_values = (sa.select(Entry.id,
                                   Entry.hash,
                                   Value.field,
                                   Value.value,
                                   File.name.label('filename'),
                                   Entry.bibkey)
                         .join_from(Entry, File).join(Value)
                         .where(sa.between(Entry.id, sa.bindparam('first'), sa.bindparam('last')))
                         .order_by('id',
                                   'field',
                                   File.priority.desc(),
                                   'filename', 'bibkey'))

        groupby_id_hash = functools.partial(itertools.groupby,
                                            key=operator.attrgetter('id', 'hash'))

        groupby_field = functools.partial(itertools.groupby,
                                          key=operator.attrgetter('field'))

        with self.connect() as conn:
            for first, last in Entry.windowed(conn, key_column='id', size=chunksize):
                result = conn.execute(select_values, {'first': first, 'last': last})
                for id_hash, grp in groupby_id_hash(result):
                    fields = [(field, [(r.value, r.filename, r.bibkey) for r in g])
                              for field, g in groupby_field(grp)]
                    yield id_hash, fields

    def merged(self, *, ref_id_field: str = REF_ID_FIELD):
        """Yield ``(bibkey, (entrytype, fields))`` entries merged by ``Entry.id``."""
        for (id_, hash_), grp in self:
            entrytype, fields = self._merged_entry(grp)
            fields[ref_id_field] = f'{id_:d}'
            yield hash_, (entrytype, fields)

    @staticmethod
    def _merged_entry(grp,
                      *, union=UNION_FIELDS, ignore=IGNORE_FIELDS,
                      raw: bool = False,
                      _sep_join=', '.join,
                      _removesuffix=_compat.removesuffix):
        # TODO: consider implementing (a subset of?) onlyifnot logic:
        # {'address': 'publisher', 'lgfamily': 'lgcode', 'publisher': 'school',
        # 'journal': 'booktitle'}
        fields = {field: values[0][0] if field not in union
                  else _sep_join(unique(vl for vl, _, _ in values))
                  for field, values in grp if field not in ignore}

        src = {_removesuffix(filename, '.bib')
               for _, values in grp
               for _, filename, _ in values}

        srctrickle = {f"{_removesuffix(filename, '.bib')}#{bibkey}"
                      for _, values in grp
                      for _, filename, bibkey in values}

        fields.update(src=_sep_join(sorted(src)),
                      srctrickle=_sep_join(sorted(srctrickle)))

        if raw:
            return fields
        entrytype = fields.pop(ENTRYTYPE)
        return entrytype, fields


class Indexable:
    """Retrieve entry from individual .bib file, or merged entry from old or new grouping."""

    def __getitem__(self, key: typing.Union[typing.Tuple[str, str], int, str]):
        """
        Entry by ``(filename, bibkey)`` or merged entry by ``refid`` (old grouping)
        or ``hash`` (current grouping).
        """
        if not isinstance(key, (tuple, int, str)):
            raise TypeError(f'key must be tuple, int, or str: {key!r}')  # pragma: no cover

        with self.connect() as conn:
            if isinstance(key, tuple):
                filename, bibkey = key
                entry = self._get_file_entry(conn, filename, bibkey)
            else:
                entry = self._get_merged_entry(conn, key=key, raw=False)

        entrytype, fields = entry
        return key, (entrytype, fields)

    @staticmethod
    def _get_file_entry(conn, filename: str, bibkey: str):
        """Return entry from individual .bib file."""
        select_items = (sa.select(Value.field,
                                  Value.value)
                        .join_from(Value, Entry)
                        .filter_by(bibkey=bibkey)
                        .join(File)
                        .filter_by(name=filename))

        result = conn.execute(select_items)
        fields = dict(iter(result))

        if not fields:
            raise KeyError((filename, bibkey))
        return fields.pop(ENTRYTYPE), fields

    @classmethod
    def _get_merged_entry(cls, conn, *, key: typing.Union[int, str],
                          raw: bool = True,
                          _get_field=operator.attrgetter('field')):
        """Return merged entry from old or new grouping."""
        keycol = Entry.refid if isinstance(key, int) else Entry.hash
        select_values = (sa.select(Value.field,
                                   Value.value,
                                   File.name.label('filename'),
                                   Entry.bibkey)
                         .join_from(Entry, File).join(Value)
                         .where(keycol == key)
                         .order_by('field',
                                   File.priority.desc(),
                                   'filename', 'bibkey'))

        result = conn.execute(select_values)
        grouped = itertools.groupby(result, key=_get_field)
        grp = [(field, [(r.value, r.filename, r.bibkey) for r in g])
               for field, g in grouped]

        if not grp:
            raise KeyError(key)
        return cls._merged_entry(grp, raw=raw)


class Exportable:
    """Write merged references into .bib file, ``bibfiles``, etc."""

    def to_bibfile(self, filepath,
                   *, encoding: str = ENCODING,
                   _sortkey: typing.Optional[str] = None):
        """Write merged references into combined .bib file at ``filepath``."""
        bibtex.save(self.merged(), str(filepath),
                    sortkey=_sortkey, encoding=encoding)

    def to_csvfile(self, filename,
                   *, dialect: str = 'excel',
                   encoding: str = ENCODING):
        """Write a CSV file with one row for each entry in each .bib file."""
        select_rows = (sa.select(File.name.label('filename'),
                                 Entry.bibkey,
                                 Entry.hash,
                                 sa.cast(Entry.id, sa.Text).label('id'))
                       .join_from(File, Entry)
                       .order_by(sa.func.lower(File.name),
                                 sa.func.lower(Entry.bibkey),
                                 'hash',
                                 Entry.id))

        with self.execute(select_rows) as result,\
             dsv.UnicodeWriter(filename, encoding=encoding, dialect=dialect) as writer:
            header = list(result.keys())
            writer.writerow(header)
            writer.writerows(result)

    def to_replacements(self, filename,
                        *, indent: typing.Optional[int] = 4):
        """Write a JSON file with 301s from merged ``glottolog_ref_id``s."""
        select_pairs = (sa.select(Entry.refid.label('id'),
                                  Entry.id.label('replacement'))
                        .where(Entry.id != Entry.refid)
                        .order_by('replacement'))

        with self.execute(select_pairs) as result:
            pairs = result.mappings().all()

        with jsonlib.update(filename, default=[], indent=indent) as repls:
            # RowMapping is not JSON serializable
            repls.extend(map(dict, pairs))

    def trickle(self, bibfiles, *, ref_id_field: str = REF_ID_FIELD):
        """Write new/changed ``glottolog_ref_id``s back into ``bibfiles``."""
        with self.connect() as conn:
            assert Entry.allid(conn=conn)

        if not self.is_uptodate(bibfiles, verbose=True):
            raise RuntimeError('trickle with an outdated db')  # pragma: no cover

        changed = (Entry.id != sa.func.coalesce(Entry.refid, -1))

        select_files = (sa.select(File.pk,
                                  File.name.label('filename'))
                        .where(sa.exists()
                               .where(Entry.file_pk == File.pk)
                               .where(changed))
                        .order_by('filename'))

        select_changed = (sa.select(Entry.bibkey,
                                    sa.cast(Entry.refid, sa.Text).label('refid'),
                                    sa.cast(Entry.id, sa.Text).label('id'))
                          .where(Entry.file_pk == sa.bindparam('file_pk'))
                          .where(changed)
                          .order_by(sa.func.lower(Entry.bibkey)))

        with self.connect() as conn:
            files = conn.execute(select_files).all()
            for file_pk, filename in files:
                bf = bibfiles[filename]
                entries = bf.load()
                added = changed = 0

                changed_entries = conn.execute(select_changed, {'file_pk': file_pk})
                for bibkey, refid, new in changed_entries:
                    entrytype, fields = entries[bibkey]
                    old = fields.pop(ref_id_field, None)
                    assert old == refid
                    if old is None:
                        added += 1
                    else:
                        changed += 1
                    fields[ref_id_field] = new

                print(f'{changed:d} changed {added:d} added in {bf.id}')
                bf.save(entries)


class Debugable:
    """Show details about splitted and combined references."""

    def stats(self, *, field_files: bool = False):
        with self.connect() as conn:
            Entry.stats(conn=conn)
            Value.fieldstats(conn=conn, with_files=field_files)
            Entry.hashstats(conn=conn)
            Entry.hashidstats(conn=conn)

    def show_splits(self):
        """Print details about bibitems that have been splitted."""
        other = sa.orm.aliased(Entry)

        select_entries = (sa.select(Entry.refid,
                                    Entry.hash,
                                    File.name.label('filename'),
                                    Entry.bibkey)
                          .join_from(Entry, File)
                          .where(sa.exists()
                                 .where(other.refid == Entry.refid)
                                 .where(other.hash != Entry.hash))
                          .order_by('refid', 'hash', 'filename', 'bibkey'))

        with self.connect() as conn:
            result = conn.execute(select_entries)
            for refid, group in groupby_first(result):
                self._print_group(conn, group)
                _, _, new = split_old_cand_new(conn, group, refid=refid)
                print(f'-> {new}\n')

    def show_merges(self):
        """Print details about bibitems that have been merged."""
        other = sa.orm.aliased(Entry)

        select_entries = (sa.select(Entry.hash,
                                    Entry.refid,
                                    File.name.label('filename'),
                                    Entry.bibkey)
                          .join_from(Entry, File)
                          .where(sa.exists()
                                 .where(other.hash == Entry.hash)
                                 .where(other.refid != Entry.refid))
                          .order_by('hash',
                                    Entry.refid.desc(),
                                    'filename', 'bibkey'))

        with self.connect() as conn:
            merge = functools.partial(merge_new_cand_old, conn,
                                      keyfunc=operator.attrgetter('refid'))
            result = conn.execute(select_entries)
            for hash_, group in groupby_first(result):
                self._print_group(conn, group)
                new, _, old = merge(group, hash_=hash_)
                print(f'-> {old}\n')

    def show_identified(self):
        """Print details about new bibitems that have been merged with present ones."""
        other = sa.orm.aliased(Entry)

        select_entries = (sa.select(Entry.hash,
                                    Entry.refid,
                                    File.name.label('filename'),
                                    Entry.bibkey)
                          .join_from(Entry, File)
                          .where(sa.exists()
                                 .where(other.refid == sa.null())
                                 .where(other.hash == Entry.hash))
                          .where(sa.exists()
                                 .where(other.refid != sa.null())
                                 .where(other.hash == Entry.hash))
                          .order_by('hash',
                                    Entry.refid != sa.null(),
                                    'refid', 'filename', 'bibkey'))

        self._show(select_entries)

    def show_combined(self):
        """Print details about new bibitems that have been merged with each other."""
        self._show_new(combined=True)

    def show_new(self):
        """Print details about new bibitems that have not been merged."""
        self._show_new(combined=False)

    def _show_new(self, *, combined: bool = False):
        other = sa.orm.aliased(Entry)

        whereclause = (sa.exists()
                       .where(other.refid == sa.null())
                       .where(other.hash == Entry.hash)
                       .where(other.pk != Entry.pk))

        if not combined:
            whereclause = ~whereclause

        select_entries = (sa.select(Entry.hash,
                                    File.name.label('filename'),
                                    Entry.bibkey)
                          .join_from(Entry, File)
                          .where(Entry.refid == sa.null())
                          .where(whereclause)
                          .order_by('hash', 'filename', 'bibkey'))

        self._show(select_entries)

    def _show(self, sql):
        with self.connect() as conn:
            result = conn.execute(sql)
            for _, group in groupby_first(result):
                self._print_group(conn, group)
                print()

    @staticmethod
    def _print_group(conn, group, *, out=print):
        for row in group:
            out(row)
        for row in group:
            hashfields = Value.hashfields(conn,
                                          filename=row.filename,
                                          bibkey=row.bibkey)
            out('\t%r, %r, %r, %r' % hashfields)


class Database(Debugable, Exportable, Indexable, BaseDatabase):
    """Collection of parsed .bib files loaded into a SQLite database."""


@registry.mapped
class File:
    """Filesystem metadata and priority setting of a .bib file."""

    __tablename__ = 'file'

    pk = sa.Column(sa.Integer, primary_key=True)

    name = sa.Column(sa.Text, sa.CheckConstraint("name != ''"), nullable=False,
                     unique=True)

    size = sa.Column(sa.Integer, sa.CheckConstraint('size > 0'), nullable=False)
    mtime = sa.Column(sa.DateTime, nullable=False)

    priority = sa.Column(sa.Integer, nullable=False)

    @classmethod
    def same_as(cls, conn, bibfiles, *, verbose: bool = False):
        """Return whether all sizes and mtimes are the same as in ``bibfiles``."""
        ondisk = {b.fname.name: (b.size, b.mtime) for b in bibfiles}

        select_files = (sa.select(cls.name, cls.size, cls.mtime)
                        .order_by('name'))
        result = conn.execute(select_files)
        indb = {name: (size, mtime) for name, size, mtime in result}

        if ondisk == indb:
            return True

        if verbose:
            ondisk_names, indb_names = (d.keys() for d in (ondisk, indb))
            print(f'missing in db: {list(ondisk_names - indb_names)}')
            print(f'missing on disk: {list(indb_names - ondisk_names)}')
            common = ondisk_names & indb_names
            differing = [c for c in common if ondisk[c] != indb[c]]
            print(f'differing in size/mtime: {differing}')
        return False


@registry.mapped
class Entry:
    """Source location and old/new grouping of a .bib entry."""

    __tablename__ = 'entry'

    pk = sa.Column(sa.Integer, primary_key=True)

    file_pk = sa.Column(sa.ForeignKey('file.pk'), nullable=False)

    bibkey = sa.Column(sa.Text, sa.CheckConstraint("bibkey != ''"),
                       nullable=False)

    # old REF_ID_FIELD from bibfiles (previous hash grouping)
    refid = sa.Column(sa.Integer, sa.CheckConstraint('refid > 0'),
                      index=True)

    # current grouping: m:n with refid (splits/merges)
    hash = sa.Column(sa.Text, sa.CheckConstraint("hash != ''"),
                     index=True)

    # split-resolved refid: every srefid maps to exactly one hash
    srefid = sa.Column(sa.Integer, sa.CheckConstraint('srefid > 0'),
                       index=True)

    # new REF_ID_FIELD to ``trickle()`` into bibfiles (current hash grouping)
    id = sa.Column(sa.Integer, sa.CheckConstraint('id > 0'),
                   index=True)

    __table_args__ = (sa.UniqueConstraint(file_pk, bibkey),)

    @classmethod
    def allhash(cls, *, conn):
        select_allhash = sa.select(~sa.exists().where(cls.hash == sa.null()))
        return conn.execute(select_allhash).scalar_one()

    @classmethod
    def allid(cls, *, conn):
        select_allid = sa.select(~sa.exists().where(cls.id == sa.null()))
        return conn.execute(select_allid).scalar_one()

    @classmethod
    def onetoone(cls, *, conn):
        other = sa.orm.aliased(cls)
        diff_id = sa.and_(other.hash == cls.hash, other.id != cls.id)
        diff_hash = sa.and_(other.id == cls.hash, other.hash != cls.hash)
        select_onetoone = sa.select(~sa.exists()
                                    .select_from(cls)
                                    .where(sa.exists()
                                           .where(sa.or_(diff_id, diff_hash))))
        return conn.execute(select_onetoone).scalar_one()

    @classmethod
    def stats(cls, *, conn, out=log.info):
        out('entry stats:')
        select_n = (sa.select(File.name.label('filename'),
                              sa.func.count().label('n'))
                    .join_from(cls, File)
                    .group_by(cls.file_pk))
        result = conn.execute(select_n)
        for r in result:
            out(f'{r.filename} {r.n:d}')

        select_total = sa.select(sa.func.count()).select_from(cls)
        total = conn.execute(select_total).scalar_one()
        out(f'{total:d} entries total')

    @classmethod
    def hashstats(cls, *, conn, out=print):
        select_total = sa.select(sa.func.count(cls.hash.distinct()).label('distinct'),
                                 sa.func.count(cls.hash).label('total'))

        result = conn.execute(select_total).one()
        out(f'{result.distinct:6d}',
            f'distinct keyids (from {result.total:d} total)', sep='\t')

        sq_1 = (sa.select(File.name.label('filename'),
                          sa.func.count(cls.hash.distinct()).label('distinct'),
                          sa.func.count(cls.hash).label('total'))
                .join_from(cls, File)
                .group_by(cls.file_pk)
                .alias())

        other = sa.orm.aliased(cls)

        sq_2 = (sa.select(File.name.label('filename'),
                          sa.func.count(cls.hash.distinct()).label('unique'))
                .join_from(cls, File)
                .where(~sa.exists()
                       .where(other.hash == cls.hash)
                       .where(other.file_pk != cls.file_pk))
                .group_by(cls.file_pk)
                .alias())

        select_files = (sa.select(sa.func.coalesce(sq_2.c.unique, 0).label('unique'),
                                  sq_1.c.filename,
                                  sq_1.c.distinct,
                                  sq_1.c.total)
                        .outerjoin_from(sq_1, sq_2, sq_1.c.filename == sq_2.c.filename)
                        .order_by(sq_1.c.filename))

        result = conn.execute(select_files)
        for r in result:
            out(f'{r.unique:6d}',
                f'{r.filename} (from {r.distinct:d} distinct'
                f' of {r.total:d} total)', sep='\t')

        select_multiple = (sa.select(sa.func.count())
                           .select_from(sa.select(sa.literal(1))
                                        .select_from(cls)
                                        .group_by(cls.hash)
                                        .having(sa.func.count(cls.file_pk.distinct()) > 1)
                                        .alias()))

        multiple = conn.execute(select_multiple).scalar_one()
        out(f'{multiple:6d}', 'in multiple files', sep='\t')

    @classmethod
    def hashidstats(cls, *, conn, out=print, ref_id_field: str = REF_ID_FIELD):
        sq = (sa.select(sa.func.count(cls.refid.distinct()).label('hash_nid'))
              .where(cls.hash != sa.null())
              .group_by(cls.hash)
              .having(sa.func.count(cls.refid.distinct()) > 1).alias())

        select_nid = (sa.select(sq.c.hash_nid,
                                sa.func.count().label('n'))
                      .group_by(sq.c.hash_nid)
                      .order_by(sa.desc('n')))

        result = conn.execute(select_nid)
        for r in result:
            out(f'1 keyid {r.hash_nid:d} {ref_id_field}s: {r.n:d}')

        sq = (sa.select(sa.func.count(cls.hash.distinct()).label('id_nhash'))
              .where(cls.refid != sa.null())
              .group_by(cls.refid)
              .having(sa.func.count(cls.hash.distinct()) > 1)
              .alias())

        select_nhash = (sa.select(sq.c.id_nhash,
                                  sa.func.count().label('n'))
                        .group_by(sq.c.id_nhash)
                        .order_by(sa.desc('n')))

        result = conn.execute(select_nhash)
        for r in result:
            out(f'1 {ref_id_field} {r.id_nhash:d} keyids: {r.n:d}')

    @classmethod
    def windowed(cls, conn, *, key_column: str, size: int):
        key_column = cls.__table__.c[key_column]
        select_keys = (sa.select(key_column.distinct())
                       .order_by(key_column))

        result = conn.execute(select_keys)
        for keys in result.scalars().partitions(size):
            yield keys[0], keys[-1]


@registry.mapped
class Value:
    """Field contents (including ENTRYTYPE) of a .bib entry."""

    __tablename__ = 'value'

    entry_pk = sa.Column(sa.ForeignKey('entry.pk'), primary_key=True)

    field = sa.Column(sa.Text, sa.CheckConstraint("field != ''"),
                      primary_key=True)

    value = sa.Column(sa.Text, nullable=False)

    __table_args__ = ({'info': {'without_rowid': True}},)

    @classmethod
    def hashfields(cls, conn, *, filename, bibkey,
                   _fields=('author', 'editor', 'year', 'title')):
        # also: extra_hash, volume (if not journal, booktitle, or series)
        select_items = (sa.select(cls.field, cls.value)
                        .where(cls.field.in_(_fields))
                        .join_from(cls, Entry)
                        .filter_by(bibkey=bibkey)
                        .join(File)
                        .filter_by(name=filename))

        fields = dict(iter(conn.execute(select_items)))
        return tuple(map(fields.get, _fields))

    @classmethod
    def fieldstats(cls, *, conn, with_files: bool = False, out=print):
        tmpl = '{n:6d}\t{field}'
        select_n = (sa.select(cls.field,
                              sa.func.count().label('n'))
                    .group_by(cls.field)
                    .order_by(sa.desc('n'), 'field'))

        if with_files:
            tmpl += '\t{files}'
            select_n = select_n.join_from(Value, Entry).join(File)
            files = sa.func.replace(sa.func.group_concat(File.name.distinct()), ',', ', ')
            select_n = select_n.add_columns(files.label('files'))

        result = conn.execute(select_n)
        for r in result.mappings():
            out(tmpl.format_map(r))


def dbapi_insert(conn, model, *, column_keys: typing.List[str],
                 executemany: bool = False,
                 paramstyle: str = 'qmark'):
    """Return callable for raw dbapi insertion of ``column_keys`` into ``model``.

    Support for ``sqlite3.Cursor.executemany(<iterator>)``.
    """
    if conn.dialect.paramstyle != paramstyle:  # pragma: no cover
        raise RuntimeError('connection dialect bad paramstyle:'
                           f' {conn.dialect.paramstyle!r}'
                           f' (require: {paramstyle!r})')

    insert_model = sa.insert(model, bind=conn)
    insert_compiled = insert_model.compile(column_keys=column_keys)

    dbapi_fairy = conn.connection
    method = dbapi_fairy.executemany if executemany else dbapi_fairy.execute
    return functools.partial(method, insert_compiled.string)


def import_bibfiles(conn, bibfiles, *, ref_id_field=REF_ID_FIELD):
    """Import bibfiles with raw dbapi."""
    log.info('importing bibfiles into a new db')

    insert_file = dbapi_insert(conn, File,
                               column_keys=['name', 'size', 'mtime', 'priority'])
    insert_entry = dbapi_insert(conn, Entry,
                                column_keys=['file_pk', 'bibkey', 'refid'])
    insert_values = dbapi_insert(conn, Value,
                                 column_keys=['entry_pk', 'field', 'value'],
                                 executemany=True)

    for b in bibfiles:
        file = (b.fname.name, b.size, b.mtime, b.priority)
        file_pk = insert_file(file).lastrowid
        for e in b.iterentries():
            entry = (file_pk, e.key, e.fields.get(ref_id_field))
            entry_pk = insert_entry(entry).lastrowid

            fields = itertools.chain([(ENTRYTYPE, e.type)], e.fields.items())
            values = ((entry_pk, field, value) for field, value in fields)
            insert_values(values)


def generate_hashes(conn):
    """Assign ``Entry.hash`` to all .bib entries for grouping."""
    from .libmonster import wrds, keyid

    words = collections.Counter()
    select_titles = (sa.select(Value.value)
                     .filter_by(field='title'))
    result = conn.execute(select_titles)
    for titles in result.scalars().partitions(10_000):
        for title in titles:
            words.update(wrds(title))
    # TODO: consider dropping stop words/hapaxes from freq. distribution
    print(f'{len(words):6d}',
          f'title words (from {sum(words.values()):d} tokens)', sep='\t')

    def windowed_entries(chunksize: int = 500):
        select_files = (sa.select(File.pk)
                        .order_by(File.name))

        files = conn.execute(select_files).scalars().all()

        select_bibkeys = (sa.select(Entry.pk)
                          .filter_by(file_pk=sa.bindparam('file_pk'))
                          .order_by('pk'))

        for file_pk in files:
            result = conn.execute(select_bibkeys, {'file_pk': file_pk})
            for entry_pks in result.scalars().partitions(chunksize):
                yield entry_pks[0], entry_pks[-1]

    select_bfv = (sa.select(Entry.pk,
                            Value.field,
                            Value.value)
                  .join_from(Value, Entry)
                  .where(Entry.pk.between(sa.bindparam('first'), sa.bindparam('last')))
                  .where(Value.field != ENTRYTYPE)
                  .order_by('pk'))

    assert conn.dialect.paramstyle == 'qmark'
    update_entry = (sa.update(Entry, bind=conn)
                    .values(hash=sa.bindparam('hash'))
                    .where(Entry.pk == sa.bindparam('entry_pk'))
                    .compile().string)
    update_entry = functools.partial(conn.connection.executemany, update_entry)

    groupby_entry_pk = functools.partial(itertools.groupby,
                                         key=operator.attrgetter('pk'))

    for first, last in windowed_entries():
        result = conn.execute(select_bfv, {'first': first, 'last': last})
        update_entry(((keyid({r.field: r.value for r in grp}, words), entry_pk)
                      for entry_pk, grp in groupby_entry_pk(result)))


def assign_ids(conn, *, verbose: bool = False):
    """Assign ``Entry.id`` to all .bib entries to establish new grouping."""
    assert Entry.allhash(conn=conn)

    reset = reset_entries(conn)
    print(f'{reset:d} entries')

    n_split = resolve_splits(conn, verbose=verbose)
    print(f'{n_split:d} splitted')

    other = sa.orm.aliased(Entry)

    no_splits = sa.select(~sa.exists()
                          .select_from(Entry)
                          .where(sa.exists()
                                 .where(other.srefid == Entry.srefid)
                                 .where(other.hash != Entry.hash)))

    assert conn.execute(no_splits).scalar_one()

    n_merged = resolve_merges(conn, verbose=verbose)
    print(f'{n_merged:d} merged')

    n_unchanged = assign_unchanged(conn)
    print(f'{n_unchanged:d} unchanged')

    no_merges = sa.select(~sa.exists()
                          .select_from(Entry)
                          .where(sa.exists()
                                 .where(other.hash == Entry.hash)
                                 .where(other.id != Entry.id)))

    assert conn.execute(no_merges).scalar_one()

    n_identified = update_identified(conn)
    print(f'{n_identified:d} identified (new/separated)')

    n_new = assign_new_and_separated(conn)
    print(f'{n_new:d} new ids (new/separated)')

    assert Entry.allid(conn=conn)
    assert Entry.onetoone(conn=conn)

    count_superseded = (sa.select(sa.func.count())
                        .where(Entry.id != Entry.srefid))

    n_superseded = conn.execute(count_superseded).scalar_one()
    print(f'{n_superseded:d} supersede pairs')


def reset_entries(conn):
    """Set srefid = refid and id = NULL on all entries."""
    update_entries = (sa.update(Entry)
                      .values(id=sa.null(),
                              srefid=Entry.refid))

    return conn.execute(update_entries).rowcount


def resolve_splits(conn, *, verbose: bool = False):
    """Keep srefid = refid only for entries from the most similar hash group."""
    other = sa.orm.aliased(Entry)

    select_split = (sa.select(Entry.refid,
                              Entry.hash,
                              File.name.label('filename'),
                              Entry.bibkey)
                    .join_from(Entry, File)
                    .where(sa.exists()
                           .where(other.refid == Entry.refid)
                           .where(other.hash != Entry.hash))
                    .order_by('refid', 'hash', 'filename', 'bibkey'))

    update_split = (sa.update(Entry)
                    .where(Entry.refid == sa.bindparam('eq_refid'))
                    .where(Entry.hash != sa.bindparam('ne_hash'))
                    .values(srefid=sa.null()))

    n_split = 0

    for refid, group in groupby_first(conn.execute(select_split)):
        n_split += len(group)

        _, _, new = split_old_cand_new(conn, group, refid=refid)

        params = {'eq_refid': refid, 'ne_hash': new}
        separated = conn.execute(update_split, params).rowcount

        if verbose:
            for row in group:
                print(row)
            for row in group:
                hashfields = Value.hashfields(conn,
                                              filename=row.filename,
                                              bibkey=row.bibkey)
                print('\t%r, %r, %r, %r' % hashfields)
            print(f'-> {new}')
            print(f'{refid:d}: {separated:d} separated from {new}\n')

    return n_split


def split_old_cand_new(conn, group, *, refid: int,
                       _get_merged_entry=Database._get_merged_entry):
    old = _get_merged_entry(conn, key=refid)

    cand = [(hash_, _get_merged_entry(conn, key=hash_))
            for hash_ in unique(r.hash for r in group)]

    new = min(cand, key=lambda p: distance(old, p[1]))[0]

    return old, cand, new


def resolve_merges(conn, *, verbose: bool = False):
    """Set id = srefid of the most similar srefid group."""
    other = sa.orm.aliased(Entry)

    select_merge = (sa.select(Entry.hash,
                              Entry.srefid,
                              File.name.label('filename'),
                              Entry.bibkey)
                    .join_from(Entry, File)
                    .where(sa.exists()
                           .where(other.hash == Entry.hash)
                           .where(other.srefid != Entry.srefid))
                    .order_by('hash',
                              Entry.srefid.desc(),
                              'filename', 'bibkey'))

    merge = functools.partial(merge_new_cand_old, conn,
                              keyfunc=operator.attrgetter('srefid'))

    update_merge = (sa.update(Entry, bind=conn)
                    .where(Entry.hash == sa.bindparam('eq_hash'))
                    .where(Entry.srefid != sa.bindparam('ne_srefid'))
                    .values(id=sa.bindparam('new_id')))

    n_merged = 0

    for hash_, group in groupby_first(conn.execute(select_merge)):
        n_merged += len(group)

        _, _, old = merge(group, hash_=hash_)

        params = {'eq_hash': hash_, 'ne_srefid': old, 'new_id': old}
        merged = conn.execute(update_merge, params).rowcount

        if verbose:
            for row in group:
                print(row)
            for row in group:
                hashfields = Value.hashfields(conn,
                                              filename=row.filename,
                                              bibkey=row.bibkey)
                print('\t%r, %r, %r, %r' % hashfields)
            print(f'-> {old}')
            print(f'{hash_}: {merged:d} merged into {old:d}\n')

    return n_merged


def merge_new_cand_old(conn, group, *, hash_: str, keyfunc,
                       _get_merged_entry=Database._get_merged_entry):
    new = _get_merged_entry(conn, key=hash_)

    cand = [(key, _get_merged_entry(conn, key=key))
            for key in unique(keyfunc(r) for r in group)]

    old = min(cand, key=lambda p: distance(new, p[1]))[0]

    return new, cand, old


def assign_unchanged(conn):
    """Set id = srefid on unchanged entries."""
    update_unchanged = (sa.update(Entry)
                        .where(Entry.id == sa.null())
                        .where(Entry.srefid != sa.null())
                        .values(id=Entry.srefid))

    return conn.execute(update_unchanged).rowcount


def update_identified(conn):
    """Set id on identified entries."""
    other = sa.orm.aliased(Entry)

    update_identified = (sa.update(Entry)
                         .where(Entry.refid == sa.null())
                         .where(sa.exists()
                                .where(other.hash == Entry.hash)
                                .where(other.id != sa.null()))
                         .values(id=(sa.select(other.id)
                                     .where(other.hash == Entry.hash)
                                     .where(other.id != sa.null())
                                     .scalar_subquery())))

    return conn.execute(update_identified).rowcount


def assign_new_and_separated(conn):
    """Assign new ids to hash groups of new/separated entries."""
    select_nextid = sa.select(sa.func.coalesce(sa.func.max(Entry.refid), 0) + 1)

    select_new = (sa.select(Entry.hash)
                  .where(Entry.id == sa.null())
                  .group_by(Entry.hash)
                  .order_by('hash'))

    assert conn.dialect.paramstyle == 'qmark'
    update_new = (sa.update(Entry, bind=conn)
                  .values(id=sa.bindparam('new_id'))
                  .where(Entry.hash == sa.bindparam('eq_hash'))
                  .compile().string)

    nextid = conn.execute(select_nextid).scalar_one()

    new_hashes = conn.execute(select_new)

    params = ((id_, hash_) for id_, (hash_,) in enumerate(new_hashes, nextid))

    dbapi_cursor = conn.connection.executemany(update_new, params)
    # https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.rowcount
    return 0 if dbapi_cursor.rowcount == -1 else dbapi_cursor.rowcount



# bibfiles.py - ordered collection of bibfiles with load/save api

import re
import collections
import unicodedata
import datetime
import math
import functools
import pathlib

import attr

from clldutils.misc import lazyproperty
from clldutils.path import memorymapped
from clldutils.source import Source
from clldutils.text import split_text
from clldutils.inifile import INI
from clldutils.attrlib import cmp_off

from . import bibtex
from . import util
from .bibfiles_db import Database

__all__ = ['BibFiles', 'BibFile', 'Entry']

ATTRS_VERSION = tuple(int(v) for v in getattr(attr, '__version__', '20.1').split('.'))

BIBFILES = 'bibfiles.sqlite3'

DOCTYPES = {k: k for k in ['grammar',
                           'grammar_sketch',
                           'dictionary',
                           'specific_feature',
                           'phonology',
                           'text',
                           'new_testament',
                           'wordlist',
                           'comparative',
                           'minimal',
                           'socling',
                           'dialectology',
                           'overview',
                           'ethnographic',
                           'bibliographical',
                           'unknown']}

PREF_YEAR_PATTERN = re.compile(r'\[(?P<year>(1|2)[0-9]{3})(\-[0-9]+)?\]')

YEAR_PATTERN = re.compile(r'(?P<year>(1|2)[0-9]{3})')


class BibFiles(list):
    """Ordered collection of `BibFile` objects accessible by filname or index."""

    @classmethod
    def from_path(cls, path, api=None):
        """BibTeX files from `<path>/bibtex/*.bib` if listed in `<path>/BIBFILES.ini`."""
        if not isinstance(path, pathlib.Path):
            path = pathlib.Path(path)
        ini = INI.from_file(path / 'BIBFILES.ini', interpolation=None)
        return cls(cls._iterbibfiles(ini, path / 'bibtex', api=api))

    @staticmethod
    def _iterbibfiles(ini, bibtex_path, api=None):
        for sec in ini.sections():
            if sec.endswith('.bib'):
                fpath = bibtex_path / sec
                if not fpath.exists():  # pragma: no cover
                    raise ValueError('invalid bibtex file referenced in BIBFILES.ini')
                yield BibFile(fname=fpath, api=api, **ini[sec])

    def __init__(self, bibfiles):
        super().__init__(bibfiles)
        self._map = {b.fname.name: b for b in self}

    def __getitem__(self, index_or_filename):
        """Retrieve a bibfile by index or filename or an entry by qualified key.

        :param index_or_filename: Either an `int` index, or a bibfile name, or a \
        provider-qualified BibTeX key in the form `<prov>:<key>`.
        :return: A `BibFile` instance, or an `Entry` instance.
        """
        if isinstance(index_or_filename, str):
            if ':' in index_or_filename:
                stem, key = index_or_filename.split(':', maxsplit=1)
                return self._map['{}.bib'.format(stem)][key]
            return self._map[index_or_filename]
        return super().__getitem__(index_or_filename)

    def to_sqlite(self, filepath=BIBFILES, rebuild=False, verbose=False):
        """Return a database with the bibfiles loaded."""
        return Database.from_bibfiles(self, filepath, rebuild=rebuild, verbose=verbose)

    def roundtrip_all(self):
        """Load and save all bibfiles with the current settings."""
        return [b.roundtrip() for b in self]


def file_if_exists(i, a, value):
    if value.exists() and not value.is_file():
        raise ValueError('invalid path')  # pragma: no cover


@attr.s
class BibFile(object):

    fname = attr.ib(validator=file_if_exists)
    name = attr.ib(default=None)
    title = attr.ib(default=None)
    description = attr.ib(default=None)
    abbr = attr.ib(default=None)
    encoding = attr.ib(default='utf-8')
    normalize = attr.ib(default='NFC')
    sortkey = attr.ib(
        default=None,
        converter=lambda s: None if s is None or s.lower() == 'none' else s)
    priority = attr.ib(default=0, converter=int)
    url = attr.ib(default=None)
    curation = attr.ib(default=None)
    api = attr.ib(default=None)

    @property
    def id(self):
        return self.fname.stem

    def __getitem__(self, item):
        if item.startswith(self.id + ':'):
            item = item.split(':', 1)[1]
        text = None
        with memorymapped(self.fname) as string:
            m = re.search(
                b'@[A-Za-z]+{' + re.escape(item.encode(self.encoding)) + rb'[\s,]', string)
            if m:
                next = string.find(b'\n@', m.end())
                if next >= 0:
                    text = string[m.start():next]
                else:
                    text = string[m.start():]
        if text:
            for k, (t, f) in bibtex.iterentries_from_text(text, encoding=self.encoding):
                return Entry(k, t, f, self, self.api)
        raise KeyError(item)

    def visit(self, visitor=None):
        entries = collections.OrderedDict()
        for entry in self.iterentries():
            if visitor is None or visitor(entry) is not True:
                entries[entry.key] = (entry.type, entry.fields)
        self.save(entries)

    @property
    def size(self):
        return self.fname.stat().st_size

    @property
    def mtime(self):
        return datetime.datetime.fromtimestamp(self.fname.stat().st_mtime)

    def iterentries(self):
        for k, (t, f) in bibtex.iterentries(filename=self.fname, encoding=self.encoding):
            yield Entry(k, t, f, self, self.api)

    def keys(self):
        return ['{0}:{1}'.format(self.id, e.key) for e in self.iterentries()]

    @property
    def glottolog_ref_id_map(self):
        return {
            e.key: e.fields['glottolog_ref_id'] for e in self.iterentries()
            if 'glottolog_ref_id' in e.fields}

    def update(self, fname, log=None):
        entries, new = collections.OrderedDict(), 0
        ref_id_map = self.glottolog_ref_id_map
        for key, (type_, fields) in bibtex.iterentries(fname, self.encoding):
            if key in ref_id_map and 'glottolog_ref_id' not in fields:
                fields['glottolog_ref_id'] = ref_id_map[key]
            else:
                new += 1
            entries[key] = (type_, fields)
        self.save(entries)
        if log:  # pragma: no cover
            log.info('{0} new entries'.format(new))

    def load(self, preserve_order=None):
        """Return entries as bibkey -> (entrytype, fields) dict."""
        if preserve_order is None:
            preserve_order = self.sortkey is None
        return bibtex.load(self.fname, preserve_order, encoding=self.encoding)

    def save(self, entries):
        """Write bibkey -> (entrytype, fields) map to file."""
        bibtex.save(
            entries,
            filename=self.fname,
            sortkey=self.sortkey,
            encoding=self.encoding,
            normalize=self.normalize)

    def __str__(self):
        return f'<{self.__class__.__name__} {self.fname.name}>'

    def check(self, log):
        entries = self.load()  # bare BibTeX syntax
        invalid = bibtex.check(filename=self.fname)  # names/macros etc.
        verdict = ('(%d invalid)' % invalid) if invalid else 'OK'
        method = log.warn if invalid else log.info
        method('%s %d %s' % (self, len(entries), verdict))
        return len(entries), verdict

    def roundtrip(self):
        print(self)
        self.save(self.load())

    def show_characters(self, include_plain=False):
        """Display character-frequencies (excluding printable ASCII)."""
        with self.fname.open(encoding=self.encoding) as fd:
            text = fd.read()
        hist = collections.Counter(text)
        table = '\n'.join(
            '%d\t%-9r\t%s\t%s' % (n, c, c, unicodedata.name(c, ''))
            for c, n in hist.most_common()
            if include_plain or not 20 <= ord(c) <= 126)
        print(table)


@functools.total_ordering
@attr.s(**cmp_off)
class Entry(object):
    key = attr.ib()
    type = attr.ib()
    fields = attr.ib()
    bib = attr.ib()
    api = attr.ib(default=None)

    # FIXME: add method to apply triggers!

    lgcode_regex = r'[a-z0-9]{4}[0-9]{4}|[a-z]{3}|NOCODE_[A-Z][^\s\]]+'
    lgcode_in_brackets_pattern = re.compile(r"\[(" + lgcode_regex + r")]")
    recomma = re.compile(r"[,/]\s?")
    lgcode_pattern = re.compile(lgcode_regex + "$")

    def __eq__(self, other):
        return self.weight == other.weight

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return self.weight < other.weight

    @property
    def _defined_doctypes(self):
        return collections.OrderedDict((hht.id, hht.id) for hht in self.api.hhtypes) \
            if self.api else DOCTYPES

    @lazyproperty
    def weight(self):
        doctypes = self._defined_doctypes
        index = len(doctypes)
        doctype = None

        for _doctype in self.doctypes(doctypes)[0]:
            index = list(doctypes.values()).index(_doctype)
            doctype = getattr(_doctype, 'id', _doctype)
            break

        # the number of pages is divided by number of doctypes times number of described languages
        pages = int(math.ceil(
            float(self.pages_int or 0) /  # noqa: W504
            ((len(self.doctypes(doctypes)[0]) or 1) *  # noqa: W504
             (len(self.lgcodes(self.fields.get('lgcode', ''))) or 1))))

        if doctype == 'grammar' and pages >= 300:
            index = -1

        return -index, pages, self.year_int or 0, self.id

    @lazyproperty
    def med_type(self):
        if self.api:
            doctypes = list(self._defined_doctypes.keys())
            index = -self.weight[0]
            if index == -1:
                return self.api.med_types.long_grammar
            if 'dictionary' in doctypes and index < doctypes.index('dictionary'):
                return self.api.med_types.get(doctypes[index])
            if 'wordlist' in doctypes and index < doctypes.index('wordlist'):
                return self.api.med_types.phonology_or_text
            return self.api.med_types.wordlist_or_less

    @lazyproperty
    def year_int(self):
        if self.fields.get('year'):
            # prefer years in brackets over the first 4-digit number.
            match = PREF_YEAR_PATTERN.search(self.fields.get('year'))
            if match:
                return int(match.group('year'))
            match = YEAR_PATTERN.search(self.fields.get('year'))
            if match:
                return int(match.group('year'))

    @lazyproperty
    def pages_int(self):
        if self.fields.get('numberofpages'):
            try:
                pages = int(self.fields.get('numberofpages').strip())
                if pages < util.MAX_PAGE:
                    return pages
            except ValueError:
                pass

        if self.fields.get('pages'):
            return util.compute_pages(self.fields['pages'])[2]

    @lazyproperty
    def publisher_and_address(self):
        p = self.fields.get('publisher')
        if p and ':' in p:
            address, publisher = [s.strip() for s in p.split(':', 1)]
            if (not self.fields.get('address')) or self.fields['address'] == address:
                return publisher, address
        return p, self.fields.get('address')

    def __str__(self):
        """Return the BibTeX representation of the entry."""
        res = "@%s{%s" % (self.type, self.key)
        for k, v in bibtex.fieldorder.itersorted(self.fields):
            res += ',\n    %s = {%s}' % (k, v.strip() if hasattr(v, 'strip') else v)
        res += '\n}\n' if self.fields else ',\n}\n'
        return res

    def text(self):
        """Return the text linearization of the entry."""
        return Source(self.type, self.key, _check_id=False, **self.fields).text()

    @property
    def id(self):
        return '{0}:{1}'.format(self.bib.id, self.key)

    @classmethod
    def lgcodes(cls, string):
        if string is None:
            return []
        codes = cls.lgcode_in_brackets_pattern.findall(string)
        if not codes:
            # ... or as comma separated list of identifiers.
            parts = [p.strip() for p in cls.recomma.split(string)]
            codes = [p for p in parts if cls.lgcode_pattern.match(p)]
            if len(codes) != len(parts):
                codes = []
        return codes

    @staticmethod
    def parse_ca(s):
        if s:
            match = re.search('computerized assignment from "(?P<trigger>[^\"]+)"', s)
            if match:
                return match.group('trigger')

    def languoids(self, langs_by_codes):
        res = []
        if 'lgcode' in self.fields:
            for code in self.lgcodes(self.fields['lgcode']):
                if code in langs_by_codes:
                    res.append(langs_by_codes[code])
        return res, self.parse_ca(self.fields.get('lgcode'))

    def doctypes(self, hhtypes):
        """Ordered doctypes assigned to this entry.

        :param hhtypes: `OrderedDict` mapping doctype names to doctypes
        :return: `list` of values of `hhtypes` which apply to the entry, ordered by occurrence in\
        `hhtypes`.
        """
        res = set()
        if 'hhtype' in self.fields:
            for ss in split_text(self.fields['hhtype'], separators=',;'):
                ss = ss.split('(')[0].strip()
                if ss in hhtypes:
                    res.add(ss)
        return [v for k, v in hhtypes.items() if k in res], self.parse_ca(self.fields.get('hhtype'))



import re
import os
import logging
import operator
import collections

from clldutils.path import move, readlines

from .languoids import Languoid, Glottocode

ISOLATE_ID = '-isolate-'
LINEAGE_SEP = ';'
NAME_AND_ID_PATTERN = re.compile(
    r'(?P<name>[^\[;]+)'
    r'(\[(?P<gc>(' + Glottocode.regex + ')|' + ISOLATE_ID + r')?])\s*'
    r'(?P<hid>[a-z]{3}|NOCODE_[^;]+)?$')


def parse_languoid(s, log):
    match = NAME_AND_ID_PATTERN.match(s.strip())
    if not match or not match.group('name').strip():
        log.error('Invalid languoid spec: {0}'.format(s))
        raise ValueError()
    return match.group('name').strip(), match.group('gc'), match.group('hid')


def rmtree(d, **kw):
    """More performant way to remove large directory structures."""
    d = str(d)
    for path in (os.path.join(d, f) for f in os.listdir(d)):
        if os.path.isdir(path):
            rmtree(path)
        else:
            os.unlink(path)
    os.rmdir(d)


def languoid(api, log, new, path, lname, glottocode, isocode, level):
    if not glottocode:
        glottocode = new.get((lname, level))
    if not glottocode:
        new[lname, level] = glottocode = api.glottocodes.new(lname)

    lineage = []
    if path:
        for i, (name, id_, hid) in enumerate(path):
            if id_ == ISOLATE_ID:
                if i != 0 or len(path) != 1:
                    log.error(
                        'invalid classification line for languoid: {0} [{1}]'.format(
                            lname, glottocode))
                    raise ValueError('invalid isolate line')
                break
            _level = api.languoid_levels.family
            if level == api.languoid_levels.dialect:
                _level = api.languoid_levels.language if i == 0 else api.languoid_levels.dialect

            if not id_:
                id_ = new.get((name, _level))
            if not id_:
                new[name, _level] = id_ = api.glottocodes.new(name)

            lineage.append((name, id_, _level, hid))

    lang = Languoid.from_name_id_level(
        api.tree, lname, glottocode, level, lineage=[(r[0], r[1], r[2]) for r in lineage], _api=api)
    if (isocode in api.iso) or (isocode is None):
        lang.iso = isocode
    lang.hid = isocode
    return lang, lineage


def read_lff(api, log, new, level, fname=None):
    assert level in [api.languoid_levels.language, api.languoid_levels.dialect]
    log.info('reading {0}s from {1}'.format(level.name, fname))

    fname = fname or api.build_path('%sff.txt' % level.name[0])

    path = None
    for line in readlines(fname):
        line = line.rstrip()
        if line.startswith('#') or not line.strip():
            # ignore comments or empty lines
            continue

        if re.match(r'\s', line):
            # leading whitespace => a language/dialect spec.
            if path is None:
                raise ValueError('language line without classification line')
            name, id_, hid = parse_languoid(line.strip(), log)
            yield languoid(api, log, new, path, name, id_, hid, level)
        else:
            path = [parse_languoid(s.strip(), log) for s in line.split(LINEAGE_SEP)]


def lang2tree(api, log, lang, lineage, out, old_tree):
    groupdir = out

    for spec in lineage:
        hid = -1
        name, id_, level = spec[:3]
        if len(spec) == 4:
            hid = spec[3]

        groupdir = groupdir.joinpath(id_)
        if not groupdir.exists():
            groupdir.mkdir()
            if id_ in old_tree:
                group = old_tree[id_]
                if group.level != level:
                    log.info('{0} from {1} to {2}'.format(group, group.level, level))
                    group.level = level
                if name != group.name:
                    # rename a subgroup!
                    group.add_name(group.name)
                    group.name = name
            else:
                group = Languoid.from_name_id_level(api.tree, name, id_, level, _api=api)

            if hid != -1:
                if (hid in api.iso or hid is None) and group.iso != hid:
                    group.iso = hid
                if hid != group.hid:
                    group.hid = hid
            group.write_info(groupdir)

    langdir = groupdir.joinpath(lang.id)
    langdir.mkdir()

    if lang.id in old_tree:
        old_lang = old_tree[lang.id]
        if old_lang.level != lang.level:
            log.info('{0} from {1} to {2}'.format(old_lang, old_lang.level, lang.level))
            old_lang.level = lang.level
        if old_lang.name != lang.name:
            old_lang.add_name(old_lang.name)
            old_lang.name = lang.name
        if old_lang.iso != lang.iso:
            old_lang.iso = lang.iso
        if lang.hid and old_lang.hid != lang.hid:
            old_lang.hid = lang.hid
        old_lang.write_info(langdir)
    else:
        lang.write_info(langdir)


def lff2tree(api, log=logging.getLogger(__name__)):
    """
    - get mapping glottocode -> Languoid from old tree
    - assemble new directory tree
      - for each path component in lff/dff:
        - create new dir
        - copy info file from old tree (possibly updating the name) or
        - create info file
      - for each language/dialect in lff/dff:
        - create new dir
        - copy info file from old tree (possibly updating the name) or
        - create info file
    - rm old tree
    - copy new tree
    """
    builddir = api.build_path('tree')
    old_tree = {lang.id: lang for lang in api.languoids()}
    out = api.tree

    if out.exists():
        if builddir.exists():
            try:
                rmtree(builddir)
            except Exception:  # pragma: no cover
                pass
            if builddir.exists():  # pragma: no cover
                raise ValueError('please remove %s before proceeding' % builddir)
        # move the old tree out of the way
        move(out, builddir)
    out.mkdir()

    new = {}
    languages = {}
    languoids = {}

    def checked(lang, lin):
        assert lang.id not in languoids
        for n, gc, _level, hid in lin:
            if gc in languoids:
                if languoids[gc] != (n, _level, hid):
                    log.error(
                        '{0}: {1} vs {2}'.format(gc, languoids[gc], (n, _level, hid)))
                    raise ValueError('inconsistent languoid data')
            else:
                languoids[gc] = (n, _level, hid)
        languoids[lang.id] = (lang.name, lang.level, lang.iso or lang.hid)
        return lang

    for lang, lineage in read_lff(
            api, log, new, api.languoid_levels.language, api.build_path('lff.txt')):
        languages[lang.id] = checked(lang, lineage)
        lang2tree(api, log, lang, lineage, out, old_tree)

    for lang, lineage in read_lff(
            api, log, new, api.languoid_levels.dialect, api.build_path('dff.txt')):
        lang = checked(lang, lineage)
        if not lang.lineage or lang.lineage[0][1] not in languages:
            log.error('missing language in dff: {0[0]} [{0[1]}]'.format(lang.lineage[0]))
            raise ValueError('invalid language referenced')

        lin = languages[lang.lineage[0][1]].lineage + lang.lineage
        lang2tree(api, log, lang, lin, out, old_tree)

    duplicates = False
    for name, getter in [('name', operator.itemgetter(0)), ('hid', operator.itemgetter(2))]:
        count = collections.Counter(getter(spec) for spec in languoids.values())
        for thing, n in count.most_common():
            if thing is None:
                continue
            if n < 2:
                break
            log.error('duplicate {0}: {1} ({2})'.format(name, thing, n))
            duplicates = True
    if duplicates:
        raise ValueError('duplicates found')


def format_comp(lang, gc=None):
    res = '{0} [{1}]'.format(lang.name, gc or lang.id)
    if lang.iso:
        res += ' {0}'.format(lang.iso)
    elif lang.hid:
        res += ' {0}'.format(lang.hid)
    return res


def format_language(lang):
    return '    {0}'.format(format_comp(lang))


def format_classification(api, lang, agg):
    if not lang.lineage:
        return format_comp(lang, gc=ISOLATE_ID)
    comps = []
    for _, gc, _ in lang.lineage:
        a = agg[gc]
        if lang.level == api.languoid_levels.language or \
                (lang.level == api.languoid_levels.dialect and  # noqa: W504
                 a.level != api.languoid_levels.family):
            comps.append(format_comp(a))
    return (LINEAGE_SEP + ' ').join(comps)


def tree2lff(api, log=logging.getLogger(__name__)):
    languoids = {api.languoid_levels.dialect: collections.defaultdict(list),
                 api.languoid_levels.language: collections.defaultdict(list)}

    agg = {}
    for lang in api.languoids():
        agg[lang.id] = lang
        if lang.level in languoids:
            languoids[lang.level][format_classification(api, lang, agg)].append(
                format_language(lang))

    for level, languages in languoids.items():
        ff = api.build_path('%sff.txt' % level.name[0])
        with ff.open('w', encoding='utf8') as fp:
            fp.write('# -*- coding: utf-8 -*-\n')
            for path in sorted(languages):
                fp.write(path + '\n')
                for lang in sorted(languages[path]):
                    fp.write(lang + '\n')
        log.info('{0}s written to {1}'.format(level.name, ff.as_posix()))



"""
Open a languoid's INI file in a text editor.
"""
import sys
import subprocess

from pyglottolog.cli_util import get_languoid


def register(parser):
    parser.add_argument(
        'languoid',
        metavar='<GLOTTOCODE>|<ISO-CODE>',
    )


def run(args):
    lang = get_languoid(args, args.languoid)
    if sys.platform.startswith('os2'):  # pragma: no cover
        cmd = 'open'
    elif sys.platform.startswith('linux'):
        cmd = 'xdg-open'
    elif sys.platform.startswith('win'):  # pragma: no cover
        cmd = 'notepad.exe'
    else:  # pragma: no cover
        print(lang.fname)
        return
    subprocess.call([cmd, str(lang.fname)])



import attr
from whoosh import index
from whoosh.fields import Schema, TEXT, KEYWORD, ID, NUMERIC
from whoosh.analysis import StemmingAnalyzer
from whoosh.qparser import QueryParser, GtLtPlugin
from whoosh.highlight import Formatter, get_text
from tqdm import tqdm

from clldutils.path import rmtree
from clldutils.misc import slug


@attr.s
class Languoid(object):
    id = attr.ib()
    iso = attr.ib()
    name = attr.ib()
    level = attr.ib()
    fname = attr.ib()
    highlights = attr.ib()


@attr.s
class Document(object):
    id = attr.ib()
    provider = attr.ib()
    title = attr.ib()
    author = attr.ib()
    authoryear = attr.ib()
    year = attr.ib()
    doctype = attr.ib()
    lgcode = attr.ib()


def get_langs_index(api, recreate=False):
    index_dir = api.build_path('whoosh_langs')
    if index_dir.exists() and recreate:
        rmtree(index_dir)  # pragma: no cover
    if not index_dir.exists():
        index_dir.mkdir()
        schema = Schema(
            id=ID(stored=True),
            name=TEXT(stored=True),
            fname=ID(stored=True),
            iso=ID(stored=True),
            level=KEYWORD(scorable=True, stored=True),
            macroarea=KEYWORD(scorable=True),
            country=KEYWORD(scorable=True),
            latitude=NUMERIC(),
            longitude=NUMERIC(),
            ini=TEXT(analyzer=StemmingAnalyzer(), stored=True),
        )
        return index.create_in(index_dir.as_posix(), schema)
    return index.open_dir(index_dir.as_posix())


class BracketFormatter(Formatter):
    """Puts square brackets around the matched terms.
    """

    def format_token(self, text, token, replace=False):
        tokentext = get_text(text, token, replace)
        return "[[%s]]" % tokentext


def search_langs(repos, q, limit=1000, **kw):
    index_ = get_langs_index(repos)
    qp = QueryParser("ini", schema=index_.schema)
    qp.add_plugin(GtLtPlugin())
    q = '{0} {1}'.format(q, ' '.join('{0}:"{1}"'.format(k, v) for k, v in kw.items()))

    def highlight(res):
        hl = res.highlights('ini', top=1)
        if hl:
            for line in hl.split('\n'):
                if '[[' in line:
                    return line.strip()

    with index_.searcher() as searcher:
        results = searcher.search(qp.parse(q), limit=limit)
        results.formatter = BracketFormatter()
        return (
            len(results),
            [
                Languoid(
                    r['id'], r.get('iso'), r['name'], r['level'], r['fname'], highlight(r)
                ) for r in results])


def build_langs_index(api, log):
    writer = get_langs_index(api, recreate=True).writer()
    for lang in tqdm(api.languoids()):
        writer.add_document(
            id=lang.id,
            name=lang.name,
            fname=str(lang.fname),
            iso=lang.iso,
            level=lang.level.name,
            macroarea=' '.join('{0}'.format(ma) for ma in lang.macroareas),
            country=' '.join('{0}'.format(c) for c in lang.countries),
            latitude=lang.latitude,
            longitude=lang.longitude,
            ini=lang.cfg.write_string(),
        )
    writer.commit()


def get_index(api, recreate=False, must_exist=False):
    index_dir = api.ftsindex
    if index_dir.exists():
        if recreate:
            rmtree(index_dir)  # pragma: no cover
    elif must_exist:
        raise ValueError('No whoosh index found at {0}.'.format(index_dir))

    if not index_dir.exists():
        index_dir.mkdir()
        schema = Schema(
            id=ID(stored=True),
            provider=KEYWORD(stored=True),
            authoryear=TEXT(stored=True),
            title=TEXT(analyzer=StemmingAnalyzer(), stored=True),
            author=TEXT(stored=True),
            year=TEXT(stored=True),
            doctype=TEXT(stored=True),
            lgcode=TEXT(stored=True),
            body=TEXT(),
            tags=KEYWORD)
        return index.create_in(index_dir.as_posix(), schema)
    return index.open_dir(index_dir.as_posix())


def search(repos, q, limit=1000, **kw):
    index_ = get_index(repos, must_exist=True)
    qp = QueryParser("body", schema=index_.schema)
    q = '{0} {1}'.format(q, ' '.join('{0}:"{1}"'.format(k, v) for k, v in kw.items()))

    with index_.searcher() as searcher:
        results = searcher.search(qp.parse(q), limit=limit)
        return len(results), [Document(**res) for res in results]


def build_index(api, log):
    writer = get_index(api, recreate=True).writer()
    for bibfile in api.bibfiles:
        log.info('indexing {0}'.format(bibfile))
        for e in tqdm(bibfile.iterentries(), leave=False):
            author = e.fields.get('author', '')
            if author:
                author = slug(author.split()[0])
            writer.add_document(
                id='{0}:{1}'.format(bibfile.fname.stem, e.key),
                provider='%s' % bibfile.fname.stem,
                title=e.fields.get('title', e.fields.get('booktitle', '')),
                author=e.fields.get('author', e.fields.get('editor', '')),
                year=e.fields.get('year', ''),
                doctype=e.fields.get('hhtype', ''),
                lgcode=e.fields.get('lgcode', ''),
                body='%s' % e.fields,
                authoryear='{0}{1}'.format(author, e.fields.get('year', '')).lower())
    writer.commit()



"""Handle references under glottolog/references."""

from .bibfiles import BibFiles, BibFile, Entry
from .hhtypes import HHTypes
from .isbns import Isbns, Isbn
from .roman import introman, romanint

__all__ = ['BibFiles', 'BibFile', 'Entry',
           'Isbns', 'Isbn',
           'HHTypes',
           'introman', 'romanint']



from collections import OrderedDict, defaultdict
import re

import attr
import markdown
import pycountry
from clldutils.misc import slug, nfilter
from clldutils import jsonlib
from dateutil import parser
import purl

from ..util import message
from ..config import AESSource, AES

__all__ = [
    'Glottocode', 'Glottocodes',
    'Reference',
    'Country',
    'ClassificationComment',
    'ISORetirement',
    'Endangerment',
    'EthnologueComment',
    'Link',
]


@attr.s(hash=True)
class Link(object):
    url = attr.ib()
    label = attr.ib(default=None)

    @property
    def domain(self):
        return purl.URL(self.url).domain()

    @classmethod
    def from_string(cls, s):
        s = s.strip()
        if s.startswith('['):
            assert s.endswith(')') and '](' in s
            return cls(*reversed(s[1:-1].split('](')))
        return cls(s)

    @classmethod
    def from_(cls, obj):
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, str):
            return cls.from_string(obj)
        if isinstance(obj, (list, tuple)) and len(obj) == 2:
            return cls(*obj)
        if isinstance(obj, dict):
            return cls(**obj)
        raise TypeError()

    def to_string(self):
        if self.label:
            return '[{0}]({1})'.format(self.label, self.url)
        return self.url

    def __json__(self):
        return attr.asdict(self)


class Glottocodes(object):
    """
    Registry keeping track of glottocodes that have been dealt out.
    """
    def __init__(self, fname):
        self._fname = fname
        self._store = jsonlib.load(self._fname)

    def __contains__(self, item):
        alpha, num = Glottocode(item).split()
        return alpha in self._store and num <= self._store[alpha]

    def __iter__(self):
        for alpha, num in self._store.items():
            for n in range(1234, num + 1):
                yield '{0}{1}'.format(alpha, n)

    def new(self, name, dry_run=False):
        alpha = slug(str(name))[:4]
        assert alpha
        while len(alpha) < 4:
            alpha += alpha[-1]
        num = self._store.get(alpha, 1233) + 1
        if not dry_run:
            self._store[alpha] = num
            # Store the updated dictionary of glottocodes back.
            ordered = OrderedDict()
            for k in sorted(self._store.keys()):
                ordered[k] = self._store[k]
            jsonlib.dump(ordered, self._fname, indent=4)
        return Glottocode('%s%s' % (alpha, num))


class Glottocode(str):
    regex = r'[a-z0-9]{4}[0-9]{4}'
    pattern = re.compile(regex + r'$')

    def __new__(cls, content):
        if not cls.pattern.match(content):
            raise ValueError(content)
        return str.__new__(cls, content)

    def split(self):
        return self[:4], int(self[4:])


@attr.s
class Reference(object):
    key = attr.ib()
    pages = attr.ib(default=None)
    trigger = attr.ib(default=None)
    pattern = re.compile(
        r"\*\*(?P<key>[a-z0-9\-_]+:[a-zA-Z.?\-;*'/()\[\]!_:0-9\u2014]+?)\*\*"
        r"(:(?P<pages>[0-9\-f]+))?"
        r'(<trigger "(?P<trigger>[^\"]+)">)?')
    old_pattern = re.compile(r'[^\[]+\[(?P<pages>[^\]]*)\]\s*\([0-9]+\s+(?P<key>[^\)]+)\)')

    def __str__(self):
        res = '**{0.key}**'.format(self)
        if self.pages:
            res += ':{0.pages}'.format(self)
        if self.trigger:
            res += '<trigger "{0.trigger}">'.format(self)
        return res

    def get_source(self, api):
        return api.bibfiles[self.bibname][self.bibkey]

    @property
    def provider(self):
        return self.key.split(':')[0]

    @property
    def bibname(self):
        return '{0}.bib'.format(self.provider)

    @property
    def bibkey(self):
        return self.key.split(':', 1)[1]

    @classmethod
    def from_match(cls, match):
        assert match
        return cls(**match.groupdict())

    @classmethod
    def from_string(cls, string, pattern=None):
        try:
            return cls.from_match((pattern or cls.pattern).match(string.strip()))
        except AssertionError:
            raise ValueError('Invalid reference: {0}'.format(string))

    @classmethod
    def from_list(cls, list_, pattern=None):
        res = []
        for s in list_:
            if s.strip():
                try:
                    res.append(cls.from_string(s, pattern=pattern))
                except AssertionError:  # pragma: no cover
                    raise ValueError('invalid ref: {0}'.format(s))
        return res


@attr.s
class Country(object):
    """
    Glottolog languoids can be related to the countries they are spoken in. These
    countries are identified by ISO 3166 Alpha-2 codes.

    .. see also:: https://en.wikipedia.org/wiki/ISO_3166-1
    """
    id = attr.ib()
    name = attr.ib()

    def __str__(self):
        return self._format()

    def _format(self, minimal=True):
        tmpl = '{0.name} ({0.id})' if not minimal else '{0.id}'
        return tmpl.format(self)

    @classmethod
    def from_name(cls, name):
        res = pycountry.countries.get(name=name)
        if res:
            return cls(id=res.alpha_2, name=res.name)

    @classmethod
    def from_id(cls, id_):
        res = pycountry.countries.get(alpha_2=id_)
        if res:
            return cls(id=res.alpha_2, name=res.name)

    @classmethod
    def from_text(cls, text):
        match = re.search(r'(?P<code_only>^[A-Z]{2}$)|\(?(?P<code>[A-Z]{2})\)?', text)
        if match:
            code = match.group('code_only') or match.group('code')
            return cls.from_id(code)
        return cls.from_name(text)


@attr.s
class ClassificationComment(object):
    sub = attr.ib(default=None)
    subrefs = attr.ib(default=attr.Factory(list), converter=Reference.from_list)
    family = attr.ib(default=None)
    familyrefs = attr.ib(default=attr.Factory(list), converter=Reference.from_list)

    def merged_refs(self, type):
        assert type in ['sub', 'family']
        res = defaultdict(set)
        for m in Reference.pattern.finditer(getattr(self, type) or ''):
            res[m.group('key')].add(m.group('pages'))
        for ref in getattr(self, type + 'refs'):
            res[ref.key].add(ref.pages)
        return [
            Reference(key=key, pages=';'.join(sorted(nfilter(pages))) or None)
            for key, pages in res.items()]

    def check(self, lang, keys, log):
        for attrib in ['subrefs', 'familyrefs']:
            for ref in getattr(self, attrib):
                if ref.key not in keys:
                    log.error(message(
                        lang, 'classification {0}: invalid bibkey: {1}'.format(attrib, ref.key)))

        for attrib in ['sub', 'family']:
            comment = getattr(self, attrib)
            if comment:
                for m in Reference.pattern.finditer(comment):
                    if m.group('key') not in keys:
                        log.error(message(
                            lang,
                            'classification {0}: invalid bibkey: {1}'.format(
                                attrib, m.group('key'))))
        return False


@attr.s
class ISORetirement(object):

    code = attr.ib(default=None)
    name = attr.ib(default=None)
    change_request = attr.ib(default=None)
    effective = attr.ib(default=None)
    reason = attr.ib(default=None)
    change_to = attr.ib(default=attr.Factory(list))
    remedy = attr.ib(default=None)
    comment = attr.ib(converter=lambda s: s.replace('\n.', '\n') if s else s, default=None)

    def asdict(self):
        return attr.asdict(self)

    __json__ = asdict


@attr.s
class Endangerment(object):
    status = attr.ib(validator=attr.validators.instance_of(AES))
    source = attr.ib(validator=attr.validators.instance_of(AESSource))
    comment = attr.ib()
    date = attr.ib(converter=parser.parse)

    def __json__(self):
        res = attr.asdict(self, recurse=True)
        res['date'] = res['date'].isoformat().split('T')[0]
        return res


def valid_ethnologue_versions(inst, attr, value):
    pattern = re.compile(r'(E[1-9][0-9]|ISO 639-3)$')
    if not all(bool(pattern.match(x)) for x in value):  # pragma: no cover
        raise ValueError('invalid ethnologue_versions: {0}'.format('/'.join(value)))


def valid_comment_type(inst, attr, value):
    if value not in ['spurious', 'missing']:
        raise ValueError('invalid comment type: {0}'.format(value))


def valid_comment(inst, attr, value):
    if not value or not isinstance(value, str):
        raise ValueError(value)


@attr.s
class EthnologueComment(object):
    # There's the isohid field which says which iso/hid the comment concerns.
    isohid = attr.ib()

    # There's the comment_type field which is either
    # - "spurious" meaning the comment is to explain why the languoid in question is
    #   spurious and in which Ethnologue (as below) that is/was
    # - "missing" meaning the comment is to explain why the languoid in question is
    #   missing (as a language entry) and in which Ethnologue (as below) that is/was
    comment_type = attr.ib(validator=valid_comment_type, converter=lambda s: s.lower())

    # There's the "ethnologue_versions" field which says which Ethnologue version(s)
    # from E16-E19 the comment pertains to, joined by /:s. E.g. E16/E17. In the case of
    # comment_type=spurious, E16/E17 in the version field means that the code was spurious
    # in E16/E17 but no longer spurious in E18/E19. In the case of comment_type=missing,
    # E16/E17 would mean that the code was missing from E16/E17, but present in E18/E19.
    # If the comment concerns a language where versions would be the empty string,
    # instead the string ISO 639-3 appears.
    ethnologue_versions = attr.ib(
        default='',
        validator=valid_ethnologue_versions,
        converter=lambda s: s.replace('693', '639').split('/'))
    comment = attr.ib(default=None, validator=valid_comment)

    def __json__(self):
        return attr.asdict(self)

    def check(self, lang, keys, log):
        try:
            markdown.markdown(self.comment)
        except Exception as e:  # pragma: no cover
            log.error(message(lang, 'ethnologue comment: invalid markup: {0}'.format(e)))
        for m in Reference.pattern.finditer(self.comment):
            if m.group('key') not in keys:
                log.error(message(lang, 'ethnologue comment: invalid bibkey: {0}'.format(
                    m.group('key'))))
        return False



"""
Create a new languoid directory for a languoid specified by name and level.
"""
import pathlib

from clldutils.clilib import ParserError

from pyglottolog.languoids import Glottocode, Languoid
from pyglottolog.cli_util import get_languoid


def register(parser):
    parser.add_argument(
        'parent',
        help='Parent languoid specified by Glottocode or directory',
    )
    parser.add_argument(
        'name',
        help='Name of the new languoid',
    )
    parser.add_argument(
        'level',
        help='Level of the languoid',
        choices=['family', 'language', 'dialect']
    )
    parser.add_argument(
        'props',
        nargs='*',
        help='Additional properties to populate the [core] section, given as prop=value pairs',
    )


def run(args):
    if Glottocode.pattern.match(args.parent):
        args.parent = get_languoid(args, args.parent).dir
    else:
        args.parent = pathlib.Path(args.parent)
        if not args.parent.exists():
            raise ParserError('invalid parent dir specified')

    lang = Languoid.from_name_id_level(
        args.parent,
        args.name,
        args.repos.glottocodes.new(args.name),
        args.level,
        **dict(prop.split('=') for prop in args.props))

    print("Info written to %s" % lang.write_info(outdir=args.parent))



"""
Fix macroarea GeoJSON
"""
import json
import collections

from clldutils.jsonlib import dump


def fixed(d, name):  # pragma: no cover
    if name in ['papunesia', 'eurasia']:
        new = None
        coordinates = []
        for polygon in d['coordinates']:
            polygon = polygon[0]

            if ([-180.0, -90.0] in polygon) or ([-180.0, 90.0] in polygon):
                # This polygon must be attached to the big one at the longitude 180 seam!
                for point in polygon:
                    point[0] = point[0] + 360
                # Now we re-order such that the polygon starts at 180,+.. and ends at 180,-90
                end, seam, start, after_bottom_left = [], [], [], False
                for point in polygon:
                    if point[0] != 180:
                        (start if after_bottom_left else end).append(point)
                    else:
                        after_bottom_left = True
                        seam.append(point)
                new = [seam[-1]] + start + end + [seam[0]]
                polygon = None
            elif ([180.0, -90.0] in polygon) or ([180.0, 90.0] in polygon):
                # The big one!
                points, in_seam = [], False
                for point in polygon:
                    if point[0] == 180.0:
                        if not in_seam:
                            points.append(point)
                            points.extend(new)
                        else:
                            print('skipping', point)
                        in_seam = True
                    else:
                        points.append(point)
                polygon = points

            if polygon:
                coordinates.append([polygon])
        d['coordinates'] = coordinates
    return d


def run(args):  # pragma: no cover
    for p in args.repos.path('config', 'macroareas', 'voronoi').glob('*.geojson'):
        with p.open(encoding='utf-8-sig') as fp:
            d = json.load(fp, object_pairs_hook=collections.OrderedDict)
        dump(fixed(d, p.stem), p, indent=4)



"""
Modules implementing functionality to add links to resources for Glottolog languoids must
subclass util.LinkProvider and override `iterupdated`, a method with the following signature:

    def iterupdated(self, languoids)

accepting the list of all Glottolog `Languoid` objects and yielding the ones that have been
updated in the function.
"""
from os.path import dirname, basename, isfile, join
import glob

modules = glob.glob(join(dirname(__file__), "*.py"))
__all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]



import collections

import requests

from .util import LinkProvider


def iter_id_iso():  # pragma: no cover
    for i in range(3):
        data = requests.get(
            "http://crubadan.org/writingsystems",
            params={
                "sEcho": "1",
                "iDisplayStart": "{0}".format(1000 * i),
                "iDisplayLength": "1000",
                "iSortingCols": "1",
                "iSortCol_0": "1",
            },
            headers={"accept": "application/json", "x-requested-with": "XMLHttpRequest"},
        ).json()
        for row in data['aaData']:
            yield (row[1].strip(), row[2].strip())


class CRUBADAN(LinkProvider):
    def iterupdated(self, languoids):  # pragma: no cover
        lmap = collections.defaultdict(list)
        for lid, iso in iter_id_iso():
            lmap[iso].append(lid)
        for lang in languoids:
            links = ["http://crubadan.org/languages/" + c for c in lmap.get(lang.iso, [])]
            if lang.update_links('crubadan.org', links):
                yield lang



"""
Map ISO codes to the list of all Glottolog languages and dialects subsumed "under" it.
"""
from csvw.dsv import UnicodeWriter

from pyglottolog.cli_util import add_output_dir


def register(parser):
    add_output_dir(parser)


def run(args):
    nodes = list(args.repos.languoids())

    res = {}
    for node in nodes:
        if node.iso:
            res[node.id] = (node.iso, set())

    for node in nodes:
        if node.level == args.repos.languoid_levels.family or node.id in res:
            continue
        for nid in res:
            matched = False
            for lineage in node.lineage:
                if lineage[1] == nid:
                    res[nid][1].add(node.id)
                    matched = True
                    break
            if matched:
                break

    with UnicodeWriter(args.output / 'iso2glottocodes.csv') as writer:
        writer.writerow(['iso', 'glottocodes'])
        for gc, (iso, gcs) in res.items():
            writer.writerow([iso, ';'.join([gc] + list(gcs))])



# Python compatibility backports

import sys

__all__ = ['nullcontext',
           'removesuffix']


if sys.version_info < (3, 7):
    import contextlib

    @contextlib.contextmanager
    def nullcontext(enter_result=None):
        yield enter_result

else:
    from contextlib import nullcontext


if sys.version_info < (3, 9):
    def removesuffix(s: str, suffix: str) -> str:
        """See https://www.python.org/dev/peps/pep-0616/

        >>> removesuffix('spam.bib', '.bib')
        'spam'
        """
        if suffix and s.endswith(suffix):
            return s[:-len(suffix)]
        else:
            return s

else:
    def removesuffix(s: str, suffix: str) -> str:
        """See https://www.python.org/dev/peps/pep-0616/

        >>> removesuffix('spam.bib', '.bib')
        'spam'
        """
        return s.removesuffix(suffix)



"""
Recreate tree from lff.txt and dff.txt
"""
from pyglottolog import lff


def run(args):
    try:
        lff.lff2tree(args.repos, args.log)
    except ValueError:  # pragma: no cover
        print("""
Something went wrong! Roll back inconsistent state running

    rm -rf languoids
    git checkout languoids
""")
        raise

    print("""
Run

    git status

to inspect changes in the directory tree.
You can run

    diff -rbB build/tree/ languoids/tree/

to inspect the changes in detail.

- To discard changes run

    git checkout languoids/tree

- To commit and push changes, run

    git add -A languoids/tree/...

  for any newly created nodes listed under

# Untracked files:
#   (use "git add <file>..." to include in what will be committed)
#
#	languoids/tree/...

  followed by

    git commit -a -m"reason for change of classification"
    git push origin
""")



from html import entities as html_entities
import pkg_resources

from nameparser import HumanName
from markdown import markdown
from clldutils.markup import iter_markdown_tables
from clldutils.jsonlib import dump


def to_html(text, url):
    c2n = html_entities.codepoint2name
    text = text.replace('&', '&amp;')
    text = ''.join('&{0};'.format(
        c2n[ord(c)]) if ord(c) > 128 and ord(c) in c2n else c for c in text)
    return markdown(text.replace(url, '[{0}]({0})'.format(url)))


def read_editions(repos):
    head, rows = next(iter_markdown_tables(
        repos.path('CONTRIBUTORS.md').read_text(encoding='utf8')))
    res = []
    for row in rows:
        row = dict(zip([c.lower() for c in head], row))
        row['editors'] = [n.strip() for n in row['editors'].split('&')]
        res.append(row)

    return sorted(res, key=lambda d: pkg_resources.parse_version(d['version']), reverse=True)


def editor_to_dict(n, editors):
    res = {'name': n}
    for e in editors.values():
        if e.name == n:
            assert e.current, '{} is not marked as current editor'.format(n)
            if e.affiliation:
                res['affiliation'] = e.affiliation
            if e.orcid:
                res['orcid'] = e.orcid
            break
    return res


def get_edition(repos, version=None):
    version = version or getattr(repos.publication.zenodo, 'version', '0.0.1.dev0')
    for edition in read_editions(repos):
        if edition['version'] == version:
            return edition
    raise ValueError('Add version {} to CONTRIBUTORS.md first!'.format(version))  # pragma: no cover


def citation(repos, edition=None, version=None):
    edition = edition or get_edition(repos, version=version)
    return "{0}. {1}. {2} {3}. {4}: {5}. (Available online at {6})".format(
        ' & '.join('{0.last}, {0.first}'.format(HumanName(e)) for e in edition['editors']),
        edition['year'],
        repos.publication.web.name,
        edition['version'],
        repos.publication.publisher.place,
        repos.publication.publisher.name,
        repos.publication.web.url,
    )


def prepare_release(repos, version=None):
    edition = get_edition(repos, version=version)
    cit = citation(repos, edition=edition, version=version)
    dump(
        {
            "title": repos.publication.zenodo.title_format.format(edition['version']),
            "description": to_html(cit, repos.publication.web.url),
            "license": {"id": repos.publication.zenodo.license_id},
            "keywords": repos.publication.zenodo.keywords.split(),
            "communities": [
                {"identifier": cid} for cid in repos.publication.zenodo.communities.split()],
            "creators": [editor_to_dict(n, repos.editors) for n in edition['editors']],
            "access_right": "open",
            "upload_type": "dataset",
        },
        repos.path('.zenodo.json'),
        indent=4)
    return cit



"""
Check the glottolog data for consistency.
"""
import collections

from clldutils.jsonlib import load
from pyglottolog.languoids import Reference
from pyglottolog.util import message
import pyglottolog.iso


def register(parser):
    parser.add_argument(
        '--bib-only',
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '--tree-only',
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '--old-languoids',
        action='store_true',
        default=False,
    )


def run(args):
    def error(obj, msg):
        args.log.error(message(obj, msg))

    def warn(obj, msg):
        args.log.warning(message(obj, msg))

    def info(obj, msg):
        args.log.info(message(obj, msg))

    if not args.tree_only:
        for bibfile in args.repos.bibfiles:
            bibfile.check(args.log)

    if args.bib_only:
        return

    refkeys = set()
    for bibfile in args.repos.bibfiles:
        refkeys = refkeys.union(bibfile.keys())

    iso = args.repos.iso
    info(iso, 'checking ISO codes')
    info(args.repos, 'checking tree')
    by_level = collections.Counter()
    by_category = collections.Counter()
    iso_in_gl, languoids, iso_splits, hid = {}, {}, [], {}
    names = collections.defaultdict(set)

    for attr in args.repos.__config__:
        for obj in getattr(args.repos, attr).values():
            ref_id = getattr(obj, 'reference_id', None)
            if ref_id and ref_id not in refkeys:
                error(obj, 'missing reference: {0}'.format(ref_id))

    old_languoids = load(args.repos.build_path('languoids.json')) if args.old_languoids else None
    old_languages_count = len(old_languoids['language']) if old_languoids else 0
    for lang in args.repos.languoids():
        # duplicate glottocodes:
        if lang.id in languoids:
            error(
                lang.id,
                'duplicate glottocode\n{0}\n{1}'.format(languoids[lang.id].dir, lang.dir))
        if old_languoids:  # pragma: no cover
            if lang.id in old_languoids[lang.level.name]:
                old_languoids[lang.level.name].remove(lang.id)
        languoids[lang.id] = lang

    if old_languoids:  # pragma: no cover
        for gc in old_languoids['language']:
            if gc not in languoids:
                error(gc, 'deleted language level languoid!')
            else:
                args.log.warning(
                    '{} switched from language to {}'.format(gc, languoids[gc].level.name))
        if not old_languoids['language']:
            args.log.info(
                'All {} language level languoids still present'.format(old_languages_count))
        for level in ['family', 'dialect']:
            if old_languoids[level]:
                args.log.info(
                    '{} {} level languoids deleted'.format(len(old_languoids[level]), level))

    for lang in languoids.values():
        ancestors = lang.ancestors_from_nodemap(languoids)
        children = lang.children_from_nodemap(languoids)

        if lang.latitude and not (-90 <= lang.latitude <= 90):
            error(lang, 'invalid latitude: {0}'.format(lang.latitude))
        if lang.longitude and not (-180 <= lang.longitude <= 180):
            error(lang, 'invalid longitude: {0}'.format(lang.longitude))

        assert isinstance(lang.countries, list)
        assert isinstance(lang.macroareas, list)
        assert (lang.timespan is None) or isinstance(lang.timespan, tuple)

        if 'sources' in lang.cfg:
            for ref in Reference.from_list(lang.cfg.getlist('sources', 'glottolog')):
                if ref.key not in refkeys:
                    error(lang, 'missing source: {0}'.format(ref))

        for attr in ['classification_comment', 'ethnologue_comment']:
            obj = getattr(lang, attr)
            if obj:
                obj.check(lang, refkeys, args.log)

        names[lang.name].add(lang)
        by_level.update([lang.level.name])
        if lang.level == args.repos.languoid_levels.language:
            by_category.update([lang.category])

        if iso and lang.iso:
            if lang.iso not in iso:
                warn(lang, 'invalid ISO-639-3 code [%s]' % lang.iso)
            else:
                isocode = iso[lang.iso]
                if lang.iso in iso_in_gl:
                    error(
                        isocode,
                        'duplicate: {0}, {1}'.format(
                            iso_in_gl[lang.iso].id, lang.id))  # pragma: no cover
                iso_in_gl[lang.iso] = lang
                isocheck = pyglottolog.iso.check_lang(
                    args.repos, isocode, lang, iso_splits=iso_splits)
                if isocheck:
                    level, lang, msg = isocheck
                    dict(info=info, warn=warn)[level](lang, msg)

        if lang.hid is not None:
            if lang.hid in hid:
                error(
                    lang.hid,
                    'duplicate hid\n{0}\n{1}'.format(languoids[hid[lang.hid]].dir, lang.dir))
            else:
                hid[lang.hid] = lang.id

        if not lang.id.startswith('unun9') and lang.id not in args.repos.glottocodes:
            error(lang, 'unregistered glottocode')
        for attr in ['level', 'name']:
            if not getattr(lang, attr):
                error(lang, 'missing %s' % attr)  # pragma: no cover
        if lang.level == args.repos.languoid_levels.language:
            parent = ancestors[-1] if ancestors else None
            if parent and parent.level != args.repos.languoid_levels.family:  # pragma: no cover
                error(lang, 'invalid nesting of language under {0}'.format(parent.level))
            for child in children:
                if child.level != args.repos.languoid_levels.dialect:  # pragma: no cover
                    error(child,
                          'invalid nesting of {0} under language'.format(child.level))
        elif lang.level == args.repos.languoid_levels.family:
            for d in lang.dir.iterdir():
                if d.is_dir():
                    break
            else:
                error(lang, 'family without children')  # pragma: no cover

        try:
            endangerment = lang.endangerment
            if endangerment and endangerment.source and endangerment.source.reference_id:
                ref = endangerment.source.reference_id
                if ref not in refkeys:  # pragma: no cover
                    error(lang, 'endangerment: invalid ref {0}'.format(ref))
        except (ValueError, KeyError) as e:  # pragma: no cover
            error(lang, 'endangerment: {0}: {1}'.format(e.__class__.__name__, str(e)))

    if iso:
        for level, obj, msg in pyglottolog.iso.check_coverage(iso, iso_in_gl, iso_splits):
            dict(info=info, warn=warn)[level](obj, msg)  # pragma: no cover

    bookkeeping_gc = args.repos.language_types.bookkeeping.pseudo_family_id
    for name, gcs in sorted(names.items()):
        if len(gcs) > 1:
            # duplicate names:
            method = error
            if len([1 for n in gcs if n.level != args.repos.languoid_levels.dialect]) <= 1:
                # at most one of the languoids is not a dialect, just warn
                method = warn  # pragma: no cover
            if len([1 for n in gcs
                    if (not n.lineage) or (n.lineage[0][1] != bookkeeping_gc)]) <= 1:
                # at most one of the languoids is not in bookkeping, just warn
                method = warn  # pragma: no cover
            method(name, 'duplicate name: {0}'.format(', '.join(sorted(
                ['{0} <{1}>'.format(n.id, n.level.name[0]) for n in gcs]))))

    def log_counter(counter, name):
        msg = [name + ':']
        maxl = max([len(k) for k in counter.keys()]) + 1
        for k, l in counter.most_common():
            msg.append(('{0:<%s} {1:>8,}' % maxl).format(k + ':', l))
        msg.append(('{0:<%s} {1:>8,}' % maxl).format('', sum(list(counter.values()))))
        print('\n'.join(msg))

    log_counter(by_level, 'Languoids by level')
    log_counter(by_category, 'Languages by category')
    return by_level



import collections

import attr
from clldutils.misc import nfilter
from clldutils.inifile import INI
from clldutils.path import Path

__all__ = [
    'AES', 'AESSource', 'Macroarea', 'DocumentType', 'LanguageType', 'LanguoidLevel',
    'Generic', 'Config']


class ConfigObject(object):
    @classmethod
    def from_section(cls, cfg, section):
        try:
            fields = set(f.name for f in attr.fields(cls))
        except attr.exceptions.NotAnAttrsClassError:
            fields = None

        kw = {'name' if 'id' in cfg[section] else 'id': section}
        kw.update(cfg[section].items())
        return cls(**{k: v for k, v in kw.items() if fields is None or k in fields})


class Generic(ConfigObject):
    def __init__(self, **kw):
        for k, v in kw.items():
            if v in ['True', 'False']:
                v = eval(v)
            setattr(self, k, v)


@attr.s
class AES(ConfigObject):
    # The attribute which is used for ordering objects of this type must come first:
    ordinal = attr.ib(converter=int)
    id = attr.ib()
    name = attr.ib()
    egids = attr.ib()
    unesco = attr.ib()
    elcat = attr.ib()
    reference_id = attr.ib()
    icon = attr.ib(default=None)


@attr.s
class AESSource(ConfigObject):
    id = attr.ib()
    name = attr.ib()
    url = attr.ib()
    reference_id = attr.ib()
    pages = attr.ib(default=None)


@attr.s
class Macroarea(ConfigObject):
    id = attr.ib()
    name = attr.ib()
    description = attr.ib()
    reference_id = attr.ib()


@attr.s
class DocumentType(ConfigObject):
    rank = attr.ib(converter=int)
    id = attr.ib()
    name = attr.ib()
    description = attr.ib()
    abbv = attr.ib()
    bibabbv = attr.ib()
    webabbr = attr.ib()
    triggers = attr.ib(converter=lambda s: nfilter(s.split('\n')))


@attr.s
class MEDType(ConfigObject):
    rank = attr.ib(converter=int)
    id = attr.ib()
    name = attr.ib()
    description = attr.ib()
    icon = attr.ib(default=None)


@attr.s
class LanguageType(ConfigObject):
    id = attr.ib()
    pseudo_family_id = attr.ib()
    category = attr.ib()
    description = attr.ib()


@attr.s(hash=True)
class LanguoidLevel(ConfigObject):
    ordinal = attr.ib(converter=int)
    id = attr.ib()
    description = attr.ib()

    @property
    def name(self):
        return self.id


def get_ini(fname, **kw):
    fname = Path(fname)
    if not fname.exists():
        # For old-style (<=3.4) repository layout we ship the config data with pyglottolog:
        name = fname.name if fname.name != 'hhtype.ini' else 'document_types.ini'
        fname = Path(__file__).parent / name
    assert fname.exists()
    return INI.from_file(fname, **kw)


class Config(collections.OrderedDict):
    __defaults__ = {}

    @classmethod
    def from_ini(cls, fname, object_class):
        ini = get_ini(fname)
        d = collections.OrderedDict()
        for sec in ini.sections():
            obj = object_class.from_section(ini, sec)
            d[obj.id] = obj
        res = cls(**d)
        res.__defaults__ = ini['DEFAULT']
        return res

    def __getattribute__(self, item):
        if item in self:
            return self[item]
        return dict.__getattribute__(self, item)

    def get(self, item, default=None):
        if isinstance(item, str) and item in self:
            return self[item]
        if isinstance(item, ConfigObject) and getattr(item, 'id', None) in self:
            return self[item.id]
        for li in self.values():
            if any(getattr(li, attr, None) == item for attr in ['name', 'value', 'description']):
                return li
        if default:
            return default
        raise ValueError(item)



"""
Update the [sources] section in languoid info files according to `lgcode` fields in bibfiles.
"""
import collections

from pyglottolog.languoids import Reference


def run(args):
    langs = args.repos.languoids_by_code()
    updated = []
    sources = collections.defaultdict(set)
    glrefs = set()

    for entry in args.repos.bibfiles['hh.bib'].iterentries():
        for lang in entry.languoids(langs)[0]:
            sources[lang.id].add('{0}:{1}'.format('hh', entry.key))
            if entry.fields.get('glottolog_ref_id'):
                glrefs.add(entry.fields['glottolog_ref_id'])

    for bib in args.repos.bibfiles:
        if bib.id == 'hh':
            continue
        for entry in bib.iterentries():
            # If language associations have already been read from an equivalent record in hh.bib,
            # we disregard the entry.
            if entry.fields.get('glottolog_ref_id') not in glrefs:
                for lang in entry.languoids(langs)[0]:
                    sources[lang.id].add('{0}:{1}'.format(bib.id, entry.key))

    for gc, refs in sources.items():
        if refs != set(r.key for r in langs[gc].sources):
            langs[gc].sources = [Reference(key=ref) for ref in sorted(refs)]
            langs[gc].write_info()
            updated.append(gc)
    print('{0} languoids updated'.format(len(updated)))



"""
Update benjamins.bib from a freshly compiled update.
"""
import argparse

from clldutils.clilib import PathType


def register(parser):
    parser.add_argument(
        'bib',
        type=PathType(type='file'),
        help='path to updated benjamins.bib')
    parser.add_argument(
        '--bibkey',
        default='benjamins.bib',
        help=argparse.SUPPRESS)


def run(args):  # pragma: no cover
    args.repos.bibfiles[args.bibkey].update(args.bib, log=args.log)



import io
import json
import pathlib
import collections

from csvw.dsv import reader, Dialect
import requests
from clldutils.misc import nfilter

from .util import LinkProvider

DOMAIN = 'collection.aiatsis.gov.au'
# The AIATSIS name data (see https://collection.aiatsis.gov.au/datasets/austlang/001)
MD_URL = "https://data.gov.au/data/dataset/70132e6f-259c-4e0f-9f95-4aed1101c053/resource/" \
         "e9a9ea06-d821-4b53-a05f-877409a1a19c/download/aiatsis_austlang_endpoint_001.csv"
# We get the mapping from AIATSIS to Glottolog from Chirila:
URL = 'http://chirila.yale.edu/languages'


class AIATSIS(LinkProvider):
    def iterupdated(self, languoids):  # pragma: no cover
        res = reader(
            io.StringIO(requests.get(MD_URL).content.decode('utf-8-sig')),
            dialect=Dialect(skipBlankRows=True, commentPrefix='<'),
            dicts=True)
        md = {d['language_code']: d for d in res}
        lmap = collections.defaultdict(set)
        return
        for line in requests.get(URL).text.splitlines():
            if line.startswith('var curItem'):
                line = line.split('=', maxsplit=1)[1]
                d = json.loads(line)
                if d['AIATSIS_Code'] and d['Glottolog_ID']:
                    codes = [c.strip().replace('*', '') for c in d['AIATSIS_Code'].split(',')]
                    for code in codes:
                        if code:
                            if code not in md:
                                print(d['AIATSIS_Code'], list(md.keys())[:10])
                                continue
                            lmap[d['Glottolog_ID']].add(code)
        with pathlib.Path(__file__).parent.joinpath('aiatsis.json').open(encoding='utf8') as fp:
            for code, gc in json.load(fp).items():
                if code not in md:
                    print(code, list(md.keys())[:10])
                    continue
                lmap[gc].add(code)

        for lang in languoids:
            links, names = [], []
            for c in sorted(lmap.get(lang.id, [])):
                links.append((md[c]['uri'], md[c]['language_name']))
                if md[c]['language_name']:
                    names.append(md[c]['language_name'])
                names.extend(nfilter([n.strip() for n in md[c]['language_synonym'].split('|')]))
            if any([
                lang.update_links(DOMAIN, links),
                lang.update_names(names, type_='aiatsis'),
            ]):
                yield lang



"""

"""
from pyglottolog.admin_commands import (
    isoretirements, langindex, updatelinks, updatemetadata, updatesources, updatemacroareas,
)


def run(args):  # pragma: no cover
    isoretirements.run(args)
    langindex.run(args)
    updatelinks.run(args)
    updatemetadata.run(args)
    updatesources.run(args)
    updatemacroareas.run(args)



import pathlib
from setuptools import setup, find_packages


setup(
    name='pyglottolog',
    version='3.5.0',
    author='Robert Forkel',
    author_email='forkel@shh.mpg.de',
    description='python package for glottolog data curation',
    long_description=pathlib.Path('README.md').read_text(encoding='utf-8'),
    long_description_content_type='text/markdown',
    keywords='data linguistics',
    license='Apache 2.0',
    url='https://github.com/clld/pyglottolog',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    platforms='any',
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'glottolog=pyglottolog.__main__:main',
            'glottolog-admin=pyglottolog.__main__:admin_main',
        ],
    },
    install_requires=[
        'gitpython',
        'pybtex>=0.22',
        'attrs>=19.2',
        'clldutils>=3.7',
        'cldfcatalog',
        'csvw>=1.5.6',
        'purl',
        'pycldf>=1.6.4',
        'sqlalchemy>=1.4',
        'tqdm',
        'latexcodec',
        'unidecode',
        'whoosh',
        'pycountry>=18.12.8',
        'termcolor',
        'newick>=0.9.2',
        'markdown',
        'requests',
        'nameparser',
    ],
    extras_require={
        'dev': ['tox>=3.14', 'flake8', 'pep8-naming', 'wheel', 'twine'],
        'test': ['pytest>=5', 'pytest-mock', 'pytest-cov'],
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



import re
import functools
import collections

import requests
import attr

from csvw.dsv import reader
from clldutils.misc import nfilter
from clldutils.attrlib import valid_range

from .util import LinkProvider
from ..languoids import Country


BASE_URL = "http://endangeredlanguages.com"
DOMAIN = "endangeredlanguages.com"
CFG_ID_NAME = "endangeredlanguages"
LINK_TYPE = "elcat"
CSV_URL = BASE_URL + "/userquery/download/"

GLOTTOCODE_MAP = {
    'afta1234': 10069, 'alab1256': 943, 'algh1238': 944, 'alip1234': 7175, 'ampi1237': 1333,
    'anei1239': 10821, 'apol1242': 1574, 'apul1237': 10922, 'ashi1243': 1577, 'assy1241': 9329,
    'auve1239': 8630, 'avok1244': 10790, 'awia1234': 2030, 'bagh1249': 1584, 'bain1259': 1419,
    'bale1254': 8030, 'bana1283': 5679, 'bana1307': 1792, 'bang1335': 1588, 'bara1408': 1123,
    'bart1238': 1592, 'barz1241': 9290, 'bear1241': 1675, 'beni1249': 283, 'bhal1244': 1599,
    'bhum1234': 1601, 'bili1250': 5561, 'binb1242': 5554, 'bodi1253': 5929, 'boka1249': 1604,
    'boon1243': 6677, 'boto1248': 10488, 'brib1244': 7176, 'brun1247': 7177, 'burg1244': 1230,
    'bwen1239': 10792, 'camb1244': 10129, 'cata1288': 10853, 'cent2328': 8554, 'chak1275': 10920,
    'chir1296': 10733, 'chon1286': 7289, 'cola1237': 6693, 'cosa1234': 10709, 'cunn1236': 10720,
    'dale1238': 3189, 'dans1239': 10452, 'dark1243': 4106, 'demu1235': 1735, 'deng1252': 2145,
    'dger1238': 1863, 'diah1239': 5134, 'dirr1240': 10465, 'djuw1238': 6229, 'dukh1234': 3196,
    'duox1238': 10625, 'duun1241': 5727, 'dyar1234': 10450, 'east2774': 8552, 'efut1241': 1456,
    'fada1248': 5931, 'farv1234': 10726, 'fore1274': 3393, 'fung1247': 547, 'fuyu1243': 3383,
    'gall1275': 3412, 'ganj1241': 4612, 'gard1245': 946, 'gari1254': 5935, 'garm1244': 10840,
    'gasc1240': 8631, 'geji1246': 2346, 'gily1242': 8556, 'girr1240': 5712, 'gube1234': 1418,
    'hach1239': 10378, 'hams1239': 10109, 'hare1244': 2442, 'hill1258': 10414, 'huzh1238': 2088,
    'hwar1238': 6038, 'hwel1241': 6106, 'ikor1238': 8194, 'indo1291': 7185, 'jaba1234': 10839,
    'jarq1234': 10074, 'jows1234': 10722, 'jude1271': 10502, 'jude1276': 10499, 'kais1242': 3214,
    'kamd1238': 5998, 'kani1276': 6688, 'kark1255': 5641, 'kart1247': 6689, 'kash1276': 10500,
    'kaur1267': 5871, 'kesh1234': 10724, 'kham1281': 8234, 'khap1242': 5884, 'khuf1238': 6108,
    'komi1277': 3280, 'komn1238': 10748, 'kuhp1234': 10073, 'kuma1284': 10475, 'lang1309': 3391,
    'lang1321': 552, 'lang1332': 10463, 'lara1261': 8116, 'lawi1235': 6045, 'lenc1243': 7256,
    'leva1238': 5657, 'limo1246': 10461, 'liya1241': 5562, 'liya1242': 5549, 'lizu1234': 10624,
    'long1406': 10814, 'lopn1238': 10481, 'lush1256': 10462, 'maan1239': 5194, 'mand1418': 5563,
    'mans1258': 8529, 'many1256': 6696, 'mard1245': 2053, 'mayy1239': 5883, 'mbiy1238': 6700,
    'meym1234': 10071, 'midd1324': 1751, 'mila1245': 1378, 'minh1238': 10471, 'miny1239': 7172,
    'mith1236': 6704, 'mpra1235': 3725, 'mudb1240': 2550, 'muya1239': 5609, 'mwen1238': 1135,
    'mwin1241': 5049, 'nasv1234': 10789, 'nati1244': 2512, 'navw1234': 3065, 'nene1249': 2062,
    'nese1235': 2089, 'ngal1294': 5546, 'ngum1253': 6717, 'nisv1234': 10791, 'niti1249': 10795,
    'norm1245': 3284, 'nort3264': 8550, 'nucl1693': 10453, 'nyan1311': 534, 'oldb1247': 7187,
    'ongk1234': 8571, 'ordo1245': 10470, 'orig1234': 7006, 'orko1234': 10801, 'oros1238': 5866,
    'pada1257': 5677, 'pail1243': 7171, 'pena1270': 8741, 'phal1255': 5938, 'phon1243': 5874,
    'phut1246': 10746, 'picu1248': 6509, 'plai1258': 9011, 'poit1240': 3269, 'prov1235': 8629,
    'pudi1238': 5451, 'ramo1243': 5684, 'rats1237': 10309, 'rucc1239': 5876, 'ruku1239': 486,
    'rush1239': 5870, 'sadu1234': 10708, 'sakh1247': 2896, 'sama1296': 5173, 'sapw1237': 10424,
    'sede1245': 10847, 'shee1239': 5932, 'shim1250': 5691, 'shir1258': 10503, 'shir1268': 5048,
    'sich1239': 5811, 'sigi1234': 10495, 'solo1263': 8570, 'sout2962': 2091, 'sout3226': 8551,
    'sout3262': 8555, 'soyo1234': 10480, 'sure1238': 5695, 'taiw1247': 2805, 'tana1291': 8118,
    'tang1377': 7170, 'tazh1234': 5180, 'tazz1244': 8553, 'teja1235': 3114, 'tets1235': 8509,
    'teus1236': 2150, 'toum1239': 5872, 'tsix1234': 4570, 'tuha1234': 10479, 'tyem1238': 5947,
    'urim1252': 10568, 'urmi1249': 9289, 'viva1235': 945, 'vivt1234': 10794, 'vovo1238': 10797,
    'wail1244': 2174, 'wang1288': 6269, 'wara1302': 2707, 'wurl1240': 5452, 'xinc1242': 9709,
    'xinc1246': 6195, 'xink1235': 6196, 'yang1305': 5560, 'yara1254': 6806, 'yaww1238': 7173,
    'yazd1240': 10501, 'yila1234': 10717, 'yogy1234': 7193, 'yulp1239': 5245, 'yuwa1242': 5987,
    'zaks1239': 10489, 'zara1252': 10444, 'zefr1234': 10072, 'zemi1238': 1253, 'zoro1244': 1630,
    'zull1239': 10464,
}


def split(s, sep=';'):
    s = re.sub('"(?P<name>[^;]+);"', lambda m: '"{}";'.format(m.group('name').strip()), s)
    return nfilter(ss.strip() for ss in s.split(sep))


def parse_coords(s):
    cc = nfilter(ss.strip().replace(' ', '') for ss in re.split('[,;]', s))
    return [Coordinate(cc[i], cc[i + 1]) for i in range(0, len(cc), 2)]


@attr.s
class Coordinate(object):
    latitude = attr.ib(converter=lambda s: float(s.strip()), validator=valid_range(-90, 90))
    longitude = attr.ib(converter=lambda s: float(s.strip()), validator=valid_range(-180, 180))


@attr.s
class ElCatLanguage(object):
    id = attr.ib(converter=int)
    isos = attr.ib(converter=functools.partial(split, sep=','))
    name = attr.ib(converter=lambda s: s.strip())
    also_known_as = attr.ib(converter=split)
    status = attr.ib()
    speakers = attr.ib()
    classification = attr.ib()
    variants_and_dialects = attr.ib(converter=split)
    u = attr.ib()
    comment = attr.ib()
    countries = attr.ib(converter=split)
    continent = attr.ib()
    coordinates = attr.ib(converter=parse_coords)

    @property
    def names(self):
        res = [self.name]
        for n in self.also_known_as:
            for nn in split(n, sep=','):
                if nn not in res:
                    res.append(nn)
        return res

    @property
    def url(self):
        return BASE_URL + '/lang/{0.id}'.format(self)


def read():
    return [ElCatLanguage(*row) for row in reader(requests.get(CSV_URL).text.split('\n')) if row]


class ElCat(LinkProvider):
    def iterupdated(self, languoids):
        elcat_langs = collections.defaultdict(list)
        for line in read():
            for iso in line.isos:
                elcat_langs[iso].append(line)
            elcat_langs[line.id].append(line)

        for lang in languoids:
            changed = False
            id_ = None
            if lang.id in GLOTTOCODE_MAP:
                if GLOTTOCODE_MAP[lang.id] in elcat_langs:
                    id_ = GLOTTOCODE_MAP[lang.id]
                else:
                    print('GLOTTOCODE_MAP: [{}]: ElCat ID {} not known'.format(
                        lang.id, GLOTTOCODE_MAP[lang.id]))
            # check for glottocode as ElCat.iso
            elif lang.id in elcat_langs:
                id_ = lang.id
            # check for ISO code as ElCat.iso
            elif lang.iso in elcat_langs:
                id_ = lang.iso
            if id_ is not None:
                if lang.update_links(
                    DOMAIN, [(l_.url, l_.name) for l_ in elcat_langs[id_]]
                ):
                    changed = True
                if len(elcat_langs[id_]) == 1:
                    # Only add alternative names, if only one ElCat language matches!
                    changed = lang.update_names(
                        elcat_langs[id_][0].names, type_=LINK_TYPE) or changed

                    # Add missing coordinates
                    if not lang.latitude:
                        # Only add missing coordinates, if ElCat lists only one coordinate pair!
                        if elcat_langs[id_][0].coordinates\
                                and len(elcat_langs[id_][0].coordinates) == 1:
                            coords = elcat_langs[id_][0].coordinates[0]
                            lang.latitude = coords.latitude
                            lang.longitude = coords.longitude
                            changed = True

                    # Add missing countries
                    if not lang.countries and elcat_langs[id_][0].countries:
                        # Only add countries, if only one ElCat language matches!
                        new_countries = nfilter([Country.from_name(c)
                                                for c in set(elcat_langs[id_][0].countries)])
                        if new_countries:
                            lang.countries = new_countries
                            changed = True

                    # Sync ElCat links and identifier
                    new_identifiers = []
                    all_elcat_link_ids = [int(li.url.split('/')[-1]) for li in lang.links
                                          if li.domain == DOMAIN]
                    for id_ in all_elcat_link_ids:
                        if id_ in elcat_langs:
                            new_identifiers.append(str(id_))
                        else:
                            print('Sync ElCat links [{}]: ElCat ID {} not known'.format(
                                lang.id, id_))
                    if new_identifiers:
                        changed = True
                        if len(new_identifiers) == 1:
                            new_identifiers = new_identifiers[0]
                        if not lang.identifier:
                            lang.cfg['identifier'] = {}
                        lang.cfg['identifier'][CFG_ID_NAME] = new_identifiers
            else:
                changed = any([lang.update_links(DOMAIN, []),
                               lang.update_names([], type_=LINK_TYPE)])

            if changed:
                yield lang

    # --- for admin purposes only ---
    def check_coverage(self, languoids):  # pragma: no cover
        elcat_langs = collections.defaultdict(list)
        elcat_ids = set()
        seen_elcat_ids = set()
        for line in read():
            for iso in line.isos:
                elcat_langs[iso].append(line)
            elcat_langs[line.id].append(line)
            elcat_ids.add(line.id)

        for lang in languoids:
            id_ = None
            if lang.id in GLOTTOCODE_MAP:
                if GLOTTOCODE_MAP[lang.id] in elcat_langs:
                    id_ = GLOTTOCODE_MAP[lang.id]
                else:
                    print('GLOTTOCODE_MAP: [{}]: ElCat ID {} not known'.format(
                        lang.id, GLOTTOCODE_MAP[lang.id]))
            # check for glottocode as ElCat.iso
            elif lang.id in elcat_langs:
                id_ = lang.id
            # check for ISO code as ElCat.iso
            elif lang.iso in elcat_langs:
                id_ = lang.iso
            if id_ is not None:
                seen_elcat_ids |= set([e.id for e in elcat_langs[id_]])

        print('Matched {} of {} ElCat languoids'.format(len(seen_elcat_ids), len(elcat_ids)))
        print('Non-matches:')
        for elcat_id in list(elcat_ids - seen_elcat_ids):
            for el in elcat_langs[elcat_id]:
                print('\t'.join(
                    [el.name, ','.join(el.isos), '{}/lang/{}'.format(BASE_URL, elcat_id)]))



"""
Write release info to .zenodo.json, CITATION.md and CONTRIBUTORS.md
"""
import git.exc
import pkg_resources

from git import Repo

from pyglottolog.metadata import prepare_release


def register(parser):
    parser.add_argument('--version', help="version number without leading 'v'", default=None)


def run(args):
    try:
        assert Repo(str(args.repos.repos)).active_branch.name.startswith('release')
    except git.exc.InvalidGitRepositoryError:
        pass
    version = getattr(args.repos.publication.zenodo, 'version', args.version)
    assert not pkg_resources.parse_version(version).is_prerelease, 'invalid release version number'
    print(prepare_release(args.repos, version=version))



"""
Search Glottolog languoids.

Note: The search index will be created upon first invocation of this command.
"""
import re
import pathlib

from termcolor import colored

from pyglottolog import fts
from pyglottolog.cli_util import register_search
from pyglottolog.util import sprint


def register(parser):
    register_search(parser, 'iso:abd')


def run(args):
    def highlight(text):
        res, i = '', 0
        for m in re.finditer(r'\[\[(?P<m>[^]]+)]]', text):
            res += text[i:m.start()]
            res += colored(m.group('m'), 'red', attrs=['bold'])
            i = m.end()
        res += text[i:]
        return res + '\n'

    count, results = fts.search_langs(args.repos, args.query)
    print('{} matches'.format(count))
    for res in results:
        try:
            p = pathlib.Path(res.fname).relative_to(pathlib.Path.cwd())
        except ValueError:
            p = res.fname
        sprint('{0.name} [{0.id}] {0.level}'.format(res), color=None, attrs=['bold'])
        sprint(p, color='green')
        sprint(highlight(res.highlights) if res.highlights else '')
    print('{} matches'.format(count))



"""
Main command line interface of the pyglottolog package.
"""
import sys
import pathlib
import argparse
import contextlib

from cldfcatalog import Config, Catalog
from clldutils.clilib import register_subcommands, get_parser_and_subparsers, ParserError
from clldutils.loglib import Logging

import pyglottolog
from pyglottolog import Glottolog
import pyglottolog.commands
import pyglottolog.admin_commands


def main(**kw):
    return _main(pyglottolog.commands, **kw)


def admin_main(**kw):
    return _main(pyglottolog.admin_commands, **kw)


def _main(commands, args=None, catch_all=False, parsed_args=None, log=None, test=False):
    try:
        repos = Config.from_file().get_clone('glottolog')
    except KeyError:  # pragma: no cover
        repos = pathlib.Path('.')
    parser, subparsers = get_parser_and_subparsers('glottolog')
    parser.add_argument(
        '--repos',
        help="clone of glottolog/glottolog",
        default=repos,
        type=pathlib.Path)
    parser.add_argument(
        '--repos-version',
        help="version of repository data. Requires a git clone!",
        default=None)
    parser.add_argument(
        '--pkg-dir',
        help=argparse.SUPPRESS,
        default=pathlib.Path(__file__).parent)
    register_subcommands(subparsers, commands)

    args = parsed_args or parser.parse_args(args=args)
    args.test = test

    if not hasattr(args, "main"):
        parser.print_help()
        return 1

    with contextlib.ExitStack() as stack:
        if not log:  # pragma: no cover
            stack.enter_context(Logging(args.log, level=args.log_level))
        else:
            args.log = log
        if args.repos_version:  # pragma: no cover
            # If a specific version of the data is to be used, we make
            # use of a Catalog as context manager:
            stack.enter_context(Catalog(args.repos, tag=args.repos_version))
        try:
            args.repos = Glottolog(args.repos)
        except Exception as e:
            print(e)
            return _main(commands, args=[args._command, '-h'])
        args.log.info('glottolog/glottolog at {0}'.format(args.repos.repos))
        try:
            return args.main(args) or 0
        except KeyboardInterrupt:  # pragma: no cover
            return 0
        except ParserError as e:
            print(e)
            return _main(commands, args=[args._command, '-h'])
        except Exception as e:  # pragma: no cover
            if catch_all:
                print(e)
                return 1
            raise


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main() or 0)
